# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import sys
import math
import numpy as np
import scipy.signal as signal
import logging

# Design notes:
# to generate an average feature file, all sweeps must have all features
# to generate a fitness score of a sweep to a feature file,, the sweep
#   must have all features in the file. If one is absent, a penalty
#   of TODO ??? will be assessed

# set of features
class EphysFeatures( object ):
    def __init__(self, name):
        # feature mean and standard deviations
        self.mean = {}
        self.stdev = {}

        # human-readable names for features
        self.glossary = {}

        # table indicating how to score feature
        #    'hit'    feature exists:
        #        'ignore'        do nothing
        #        'stdev'            score is # stdevs from target mean
        #    'miss'    feature absent:
        #        'constant'        score = scoring['constant']
        #        'mean_mult'        score = mean * scoring['mean_mult']
        #
        self.scoring = {}

        self.name = name

                ################################################################
        # ignore scores
        ignore_score = { "hit": "ignore" }
        self.glossary["n_spikes"] = "Number of spikes"
        self.scoring["n_spikes"] = ignore_score

        ################################################################
        # ignore misses
        ignore_miss = { "hit":"stdev", "miss":"const", "const":0 }
        self.glossary["adapt"] = "Adaptation index"
        self.scoring["adapt"] = ignore_miss
        self.glossary["latency"] = "Time to first spike (ms)"
        self.scoring["latency"] = ignore_miss

        ################################################################
        # base miss off mean
        mean_score = { "hit":"stdev", "miss":"mean_mult", "mean_mult":2 }
        self.glossary["ISICV"] = "ISI-CV"
        self.scoring["ISICV"] = mean_score

        ################################################################
        # normal scoring
        normal_score = { "hit":"stdev", "miss":"const", "const":20 }
        self.glossary["isi_avg"] = "Average ISI (ms)"
        self.scoring["isi_avg"] = ignore_score
        self.glossary["doublet"] = "Doublet ISI (ms)"
        self.scoring["doublet"] = normal_score
        self.glossary["f_fast_ahp"] = "Fast AHP (mV)"
        self.scoring["f_fast_ahp"] = normal_score
        self.glossary["f_slow_ahp"] = "Slow AHP (mV)"
        self.scoring["f_slow_ahp"] = normal_score
        self.glossary["f_slow_ahp_time"] = "Slow AHP time"
        self.scoring["f_slow_ahp_time"] = normal_score
        self.glossary["base_v"] = "Baseline voltage (mV)"
        self.scoring["base_v"] = normal_score
        #self.glossary["base_v2"] = "Baseline voltage 2 (mV)"
        #self.scoring["base_v2"] = normal_score
        #self.glossary["base_v3"] = "Baseline voltage 3 (mV)"
        #self.scoring["base_v3"] = normal_score
        ################################################################
        # per spike scoring
        perspike_score = { "hit":"perspike", "miss":"const", "const":20, "skip_last_n":0 }
        self.glossary["f_peak"] = "Spike height (mV)"
        self.scoring["f_peak"] = perspike_score.copy()
        self.glossary["f_trough"] = "Spike depth (mV)"
        self.scoring["f_trough"] = perspike_score.copy()
        self.scoring["f_trough"]["skip_last_n"] = 1
        # self.glossary["f_w"] = "Spike width at -30 mV (ms)"
        # self.scoring["f_w"] = perspike_score.copy()
        self.glossary["upstroke"] = "Peak upstroke (mV/ms)"
        self.scoring["upstroke"] = perspike_score.copy()
        self.glossary["upstroke_v"] = "Vm of peak upstroke (mV)"
        self.scoring["upstroke_v"] = perspike_score.copy()
        self.glossary["downstroke"] = "Peak downstroke (mV/ms)"
        self.scoring["downstroke"] = perspike_score.copy()
        self.glossary["downstroke_v"] = "Vm of peak downstroke (mV)"
        self.scoring["downstroke_v"] = perspike_score.copy()
        self.glossary["threshold"] = "Threshold voltage (mV)"
        self.scoring["threshold"] = perspike_score.copy()
        self.glossary["width"] = "Spike width at half-max (ms)"
        self.scoring["width"] = perspike_score.copy()
        self.scoring["width"]["skip_last_n"] = 1
        self.glossary["thresh_ramp"] = "Change in dv/dt over first 5 mV past threshold (mV/ms)"
        self.scoring["thresh_ramp"] = perspike_score.copy()


        ################################################################
        # heavily penalize when there are no spikes
        spike_score = { "hit":"stdev", "miss":"const", "const":250 }
        self.glossary["rate"] = "Firing rate (Hz)"
        self.scoring["rate"] = spike_score

    def print_out(self):
        print("Features from " + self.name)
        for k in self.mean.keys():
            if k in self.glossary:
                st = "%30s = " % self.glossary[k]
                if self.mean[k] is not None:
                    st += "%g" % self.mean[k]
                else:
                    st += "--------"
                if k in self.stdev and self.stdev[k] is not None:
                    st += " +/- %g" % self.stdev[k]
                print(st)

    # initialize summary feature set from file
    def clone(self, param_dict):
        for k in param_dict.keys():
            self.mean[k] = param_dict[k]["mean"]
            self.stdev[k] = param_dict[k]["stdev"]

class EphysFeatureExtractor( object ):
    def __init__(self):
        # list of feature set instances
        self.feature_list = []
        # names of each element in feature list
        self.feature_source = []
        # feature set object representing combination of all instances
        self.summary = None

    # adds new feature set instance to feature_list
    def process_instance(self, name, v, curr, t, onset, dur, stim_name):
        feature = EphysFeatures(name)

        ################################################################
        # set stop time -- run until end of stimulus or end of sweep
        # comment-out the one of the two approaches
        # detect spikes only during stimulus
        start = onset
        stop = onset + dur
        # detect spikes for all of sweep
        #start = 0
        #stop = t[-1]
        ################################################################
        # pull out spike times

        # calculate the derivative only within target window
        # otherwise get spurious detection at ends of stimuli
        # filter with 10kHz cutoff if constant 200kHz sample rate (ie experimental trace)
        start_idx = np.where(t >= start)[0][0]
        stop_idx = np.where(t >= stop)[0][0]
        v_target = v[start_idx:stop_idx]
        if np.abs(t[1] - t[0] - 5e-6) < 1e-7 and np.var(np.diff(t)) < 1e-6:
            b, a = signal.bessel(4, 0.1, "low")
            smooth_v = signal.filtfilt(b, a, v_target, axis=0)
            dv = np.diff(smooth_v)
        else:
            dv = np.diff(v_target)
        dvdt = dv / (np.diff(t[start_idx:stop_idx]) * 1e3) # in mV/ms

        dv_cutoff = 20
        thresh_pct = 0.05
        spikes = []
        temp_spk_idxs = np.where(np.diff(np.greater_equal(dvdt, dv_cutoff).astype(int)) == 1)[0] # find positive-going crossings of 100 mV/ms
        spk_idxs = []
        for i, temp in enumerate(temp_spk_idxs):
            if i == 0:
                spk_idxs.append(temp)
            elif np.any(dvdt[temp_spk_idxs[i - 1]:temp] < 0):
                # check if the dvdt has gone back down below zero between presumed spike times
                # sometimes the dvdt bobbles around detection threshold and produces spurious guesses at spike times
                spk_idxs.append(temp)
        spk_idxs += start_idx # set back to the "index space" of the original trace

        # recalculate full dv/dt for feature analysis (vs spike detection)
        if np.abs(t[1] - t[0] - 5e-6) < 1e-7 and np.var(np.diff(t)) < 1e-6:
            b, a = signal.bessel(4, 0.1, "low")
            smooth_v = signal.filtfilt(b, a, v, axis=0)
            dv = np.diff(smooth_v)
        else:
            dv = np.diff(v)
        dvdt = dv / (np.diff(t) * 1e3) # in mV/ms

        # First time through, accumulate upstrokes to calculate average threshold target
        for spk_n, spk_idx in enumerate(spk_idxs):
            # Etay defines spike as time of threshold crossing
            spk = {}

            if spk_n < len(spk_idxs) - 1:
                next_idx = spk_idxs[spk_n + 1]
            else:
                next_idx = stop_idx

            if spk_n > 0:
                prev_idx = spk_idxs[spk_n - 1]
            else:
                prev_idx = start_idx

            # Find the peak
            peak_idx = np.argmax(v[spk_idx:next_idx]) + spk_idx

            spk["peak_idx"] = peak_idx
            spk["f_peak"] = v[peak_idx]
            spk["f_peak_i"] = curr[peak_idx]
            spk["f_peak_t"] = t[peak_idx]

            # Check if end of stimulus interval cuts off spike - if so, don't process spike
            if spk_n == len(spk_idxs) - 1 and peak_idx == next_idx-1:
                continue
            if spk_idx == peak_idx:
                continue    # this was bugfix, but why? ramp?

            # Determine maximum upstroke of spike
            upstroke_idx = np.argmax(dvdt[spk_idx:peak_idx]) + spk_idx

            spk["upstroke"] = dvdt[upstroke_idx]
            if np.isnan(spk["upstroke"]): # sometimes dvdt will be NaN because of multiple cvode points at same time step
                close_idx = upstroke_idx + 1
                while (np.isnan(dvdt[close_idx])):
                    close_idx += 1
                spk["upstroke_idx"] = close_idx
                spk["upstroke"] = dvdt[close_idx]
                spk["upstroke_v"] = v[close_idx]
                spk["upstroke_i"] = curr[close_idx]
                spk["upstroke_t"] = t[close_idx]
            else:
                spk["upstroke_idx"] = upstroke_idx
                spk["upstroke_v"] = v[upstroke_idx]
                spk["upstroke_i"] = curr[upstroke_idx]
                spk["upstroke_t"] = t[upstroke_idx]

            # Preliminarily define threshold where dvdt = 5% * max upstroke
            thresh_pct = 0.05
            find_thresh_idxs = np.where(dvdt[prev_idx:upstroke_idx] <= thresh_pct * spk["upstroke"])[0]
            if len(find_thresh_idxs) < 1: # Can't find a good threshold value - probably a bad simulation case
                # Fall back to the upstroke value
                threshold_idx = upstroke_idx
            else:
                threshold_idx = find_thresh_idxs[-1] + prev_idx
            spk["threshold_idx"] = threshold_idx
            spk["threshold"] = v[threshold_idx]
            spk["threshold_v"] = v[threshold_idx]
            spk["threshold_i"] = curr[threshold_idx]
            spk["threshold_t"] = t[threshold_idx]
            spk["rise_time"] = spk["f_peak_t"] - spk["threshold_t"]

            PERIOD = t[1] - t[0]
            width_volts = (v[peak_idx] + v[threshold_idx]) / 2
            recording_width = False
            for i in range(threshold_idx, min(len(v), threshold_idx + int(0.001 / PERIOD))):
                if not recording_width and v[i] >= width_volts:
                    recording_width = True
                    idx0 = i
                elif recording_width and v[i] < width_volts:
                    spk["half_height_width"] = t[i] - t[idx0]
                    break
            # </KEITH>

            # Check for things that are probably not spikes:
            # if there is more than 2 ms between the detection event and the peak, don't count it
            if t[peak_idx] - t[threshold_idx] > 0.002:
                continue
            # if the "spike" is less than 2 mV, don't count it
            if v[peak_idx] - v[threshold_idx] < 2.0:
                continue
            # if the absolute value of the peak is less than -30 mV, don't count it
            if v[peak_idx] < -30.0:
                continue
            spikes.append(spk)

        # Refine threshold target based on average of all spikes
        if len(spikes) > 0:
            threshold_target = np.array([spk["upstroke"] for spk in spikes]).mean() * thresh_pct

        for spk_n, spk in enumerate(spikes):
            if spk_n < len(spikes) - 1:
                next_idx = spikes[spk_n + 1]["threshold_idx"]
            else:
                next_idx = stop_idx

            if spk_n > 0:
                prev_idx = spikes[spk_n - 1]["peak_idx"]
            else:
                prev_idx = start_idx

            # Restore variables from before
            # peak_idx = spk['peak_idx']
            peak_idx = np.argmax(v[spk['threshold_idx']:next_idx]) + spk['threshold_idx']

            spk["peak_idx"] = peak_idx
            spk["f_peak"] = v[peak_idx]
            spk["f_peak_i"] = curr[peak_idx]
            spk["f_peak_t"] = t[peak_idx]

            # Determine maximum upstroke of spike
            # upstroke_idx = spk['upstroke_idx']
            upstroke_idx = np.argmax(dvdt[spk['threshold_idx']:peak_idx]) + spk['threshold_idx']

            spk["upstroke"] = dvdt[upstroke_idx]
            if np.isnan(spk["upstroke"]): # sometimes dvdt will be NaN because of multiple cvode points at same time step
                close_idx = upstroke_idx + 1
                while (np.isnan(dvdt[close_idx])):
                    close_idx += 1
                spk["upstroke_idx"] = close_idx
                spk["upstroke"] = dvdt[close_idx]
                spk["upstroke_v"] = v[close_idx]
                spk["upstroke_i"] = curr[close_idx]
                spk["upstroke_t"] = t[close_idx]
            else:
                spk["upstroke_idx"] = upstroke_idx
                spk["upstroke_v"] = v[upstroke_idx]
                spk["upstroke_i"] = curr[upstroke_idx]
                spk["upstroke_t"] = t[upstroke_idx]

            # Find threshold based on average target
            find_thresh_idxs = np.where(dvdt[prev_idx:upstroke_idx] <= threshold_target)[0]
            if len(find_thresh_idxs) < 1: # Can't find a good threshold value - probably a bad simulation case
                # Fall back to the upstroke value
                threshold_idx = upstroke_idx
            else:
                threshold_idx = find_thresh_idxs[-1] + prev_idx
            spk["threshold_idx"] = threshold_idx
            spk["threshold"] = v[threshold_idx]
            spk["threshold_v"] = v[threshold_idx]
            spk["threshold_i"] = curr[threshold_idx]
            spk["threshold_t"] = t[threshold_idx]

            # Define the spike time as threshold time
            spk["t_idx"] = threshold_idx
            spk["t"] = t[threshold_idx]

            # Save the -30 mV crossing time for backward compatibility with Etay code
            overn30_idxs = np.where(v[threshold_idx:peak_idx] >= -30)[0]
            if len(overn30_idxs) > 0:
                spk["t_idx_n30"] = overn30_idxs[0] + threshold_idx
            else: # fall back to threshold definition if spike doesn't cross -30 mV
                spk["t_idx_n30"] = threshold_idx
            spk["t_n30"] = t[spk["t_idx_n30"]]

            # Figure out initial "slope" of phase plot post-threshold
            plus_5_vec = np.where(v[threshold_idx:upstroke_idx] >= spk["threshold"] + 5)[0]
            if len(plus_5_vec) > 0:
                thresh_plus_5_idx = plus_5_vec[0] + threshold_idx
                spk["thresh_ramp"] = dvdt[thresh_plus_5_idx] - dvdt[threshold_idx]
            else:
                spk["thresh_ramp"] = dvdt[upstroke_idx] - dvdt[threshold_idx]

            # go forward to determine peak downstroke of spike
            downstroke_idx = np.argmin(dvdt[peak_idx:next_idx]) + peak_idx
            spk["downstroke_idx"] = downstroke_idx
            spk["downstroke_v"] = v[downstroke_idx]
            spk["downstroke_i"] = curr[downstroke_idx]
            spk["downstroke_t"] = t[downstroke_idx]
            spk["downstroke"] = dvdt[downstroke_idx]
            if np.isnan(spk["downstroke"]): # sometimes dvdt will be NaN because of multiple cvode points at same time step
                close_idx = downstroke_idx + 1
                while (np.isnan(dvdt[close_idx])):
                    close_idx += 1
                spk["downstroke"] = dvdt[close_idx]

        features = {}
        feature.mean["base_v"] = v[np.where((t > onset - 0.1) & (t < onset - 0.001))].mean() # baseline voltage, 100ms before stim
        feature.mean["spikes"] = spikes
        isi_cv = self.isicv(spikes)
        if isi_cv is not None:
            feature.mean["ISICV"] = isi_cv
        n_spikes = len(spikes)
        feature.mean["n_spikes"] = n_spikes
        feature.mean["rate"] = 1.0 * n_spikes / (stop - start);
        feature.mean["adapt"] = self.adaptation_index(spikes, stop)
        if len(spikes) > 1:
            feature.mean["doublet"] = 1000 * (spikes[1]["t"] - spikes[0]["t"])
        if len(spikes) > 0:
            for i, spk in enumerate(spikes):
                idx_next = spikes[i + 1]["t_idx"] if i < len(spikes) - 1 else stop_idx
                self.calculate_trough(spk, v, curr, t, idx_next)
                half_max_v = (spk["f_peak"] - spk["f_trough"]) / 2.0 + spk["f_trough"]
                over_half_max_v_idxs = np.where(v[spk["t_idx"]:spk["trough_idx"]] > half_max_v)[0]
                if len(over_half_max_v_idxs) > 0:
                    spk["width"] = 1000. * (t[over_half_max_v_idxs[-1] + spk["t_idx"]] - t[over_half_max_v_idxs[0] + spk["t_idx"]])
            feature.mean["latency"] = 1000. * (spikes[0]["t"] - onset)
            feature.mean["latency_n30"] = 1000. * (spikes[0]["t_n30"] - onset)
            # extract properties for each spike
            isicnt = 0
            isitot = 0
            for i in range(0, len(spikes)-1):
                spk = spikes[i]
                idx_next = spikes[i+1]["t_idx"]
                isitot += spikes[i+1]["t"] - spikes[i]["t"]
                isicnt += 1
            if isicnt > 0:
                feature.mean["isi_avg"] = 1000 * isitot / isicnt
            else:
                feature.mean["isi_avg"] = None
        # average feature data from individual spikes
        # build superset dictionary of possible features
        superset = {}
        for i in range(len(spikes)):
            for k in spikes[i].keys():
                if k not in superset:
                    superset[k] = k

        for k in superset.keys():
            cnt = 0
            mean = 0
            for i in range(len(spikes)):
                if k not in spikes[i]:
                    continue
                mean += float(spikes[i][k])
                cnt += 1.0
            # this shouldn't be possible, but it may be in future version
            #   so might as well trap for it
            if cnt == 0:
                continue
            mean /= cnt
            stdev = 0
            for i in range(len(spikes)):
                if k not in spikes[i]:
                    continue
                dif = mean - float(spikes[i][k])
                stdev += dif * dif
            stdev = math.sqrt(stdev / cnt)
            feature.mean[k] = mean
            feature.stdev[k] = stdev
        #
        self.feature_list.append(feature)
        self.feature_source.append(name)

    def isicv(self, spikes):
        if len(spikes) < 3:
            return None
        isi_mean = 0
        lst = []
        for i in range(len(spikes) - 1):
            isi = spikes[i+1]["t"] - spikes[i]["t"]
            #print("\t%g" % isi)
            isi_mean += isi
            lst.append(isi)
        isi_mean /= 1.0 * len(lst)
        #print(isi_mean)
        var = 0
        for i in range(len(lst)):
            dif = isi_mean - lst[i]
            var += dif * dif
        var /= len(lst)
        #var /= len(lst) - 1
        #print(math.sqrt(var))
        if isi_mean > 0:
            return math.sqrt(var) / isi_mean
        return None

    def adaptation_index(self, spikes, stim_end):
        if len(spikes) < 4:
            return None
        adi = 0
        cnt = 0
        isi = []
        for i in range(len(spikes)-1):
            isi.append(spikes[i+1]["t"] - spikes[i]["t"])
        # act as though time between last spike and stim end is another ISI per Etay's code
        # l = stim_end - spikes[-1]["t"]
        # if l > 0 and l > isi[-1]:
        #     isi.append(l)
        for i in range(len(isi)-1):
            adi += 1.0 * (isi[i+1] - isi[i]) / (isi[i+1] + isi[i])
            cnt += 1
        adi /= cnt
        return adi

    ##----------------------------------------------------------------------

    # trough (AHP) is presently defined as the minimum voltage level
    #   observed between successive spikes in a burst
    # there's too much data to cleanly return it on the stack
    # instead, spike table is passed in instead
    def calculate_trough(self, spike, v, curr, t, next_idx):
        # dt = t[1] - t[0]
        peak_idx = spike["peak_idx"]

        if peak_idx >= next_idx:
            logging.warning("next index (%d) before peak index (%d) calculating trough" % ( next_idx, peak_idx ))
            trough_idx = next_idx
        else:
            trough_idx = np.argmin(v[peak_idx:next_idx]) + peak_idx

        spike["trough_idx"] = trough_idx
        spike["f_trough"] = v[trough_idx]
        spike["trough_v"] = v[trough_idx]
        spike["trough_t"] = t[trough_idx]
        spike["trough_i"] = curr[trough_idx]

        # calculate etay's 'fast' and 'slow' ahp here
        if t[peak_idx] + 0.005 >= t[-1]:
            five_ms_idx = len(t) - 1
        else:
            five_ms_idx = np.where(t >= 0.005 + t[peak_idx])[0][0]  # 5ms after peak

        # fast AHP is minimum value occurring w/in 5ms
        if five_ms_idx >= next_idx:
            five_ms_idx = next_idx

        if peak_idx == five_ms_idx:
            fast_idx = next_idx
        else:
            fast_idx = np.argmin(v[peak_idx:five_ms_idx]) + peak_idx

        spike["f_fast_ahp"] = v[fast_idx]
        spike["f_fast_ahp_v"] = v[fast_idx]
        spike["f_fast_ahp_i"] = curr[fast_idx]
        spike["f_fast_ahp_t"] = t[fast_idx]

        if five_ms_idx == next_idx:
            slow_idx = fast_idx
        else:
            slow_idx = np.argmin(v[five_ms_idx:next_idx]) + five_ms_idx

        spike["f_slow_ahp"] = v[slow_idx]
        spike["f_slow_ahp_time"] = (t[slow_idx] - t[peak_idx]) / (t[next_idx] - t[peak_idx])
        spike["f_slow_ahp_t"] = t[slow_idx]

    # initialize summary feature set from file
    def push_summary(self, new_summary):
        self.summary = new_summary

    # calculate nearness score for feature set X relative to summary
    # the nearness score is the sum of squares the features are from
    #   their target values, in units of standard deviations
    # when a feature is absent, the algorithm to determine the
    #   penalty is stored in the feature itself, and this value
    #   is calculated then added to the sum
    def score_feature_set(self, set_num):
        cand = self.feature_list[set_num]
        scores = []
        for k in sorted(self.summary.mean.keys()):
            if k in self.summary.glossary:
                response = self.summary.scoring[k]["hit"]
                if response == "ignore":
                    continue
                elif response == "stdev":
                    mean = self.summary.mean[k]
                    stdev = self.summary.stdev[k]
                    assert stdev > 0
                    if k in cand.mean and cand.mean[k] is not None:
                        val = cand.mean[k]
                        inc = abs(mean - val) / stdev
                        scores.append(inc)
#                        print("Hit %s, %g+/-%g (%g) = %g" % (k, mean, stdev, val, inc))
                    else:
                        resp = cand.scoring[k]["miss"]
                        if resp == "const":
                            miss = float(cand.scoring[k][resp])
                        elif resp == "mean_mult":
                            miss = mean * float(cand.scoring[k][resp])
                        else:
                            assert False
                        miss = float(miss)
                        scores.append(miss)
#                        print("Missed %s, penalty = %g" % (k, miss))
                elif response == "perspike":
                    mean = self.summary.mean[k]
                    stdev = self.summary.stdev[k]
                    assert stdev > 0
                    if k in cand.mean and cand.mean[k] is not None:
                        val = 0
                        n_spikes = len(cand.mean["spikes"])
                        skip_last_n = self.summary.scoring[k]["skip_last_n"]
                        for spike in cand.mean["spikes"][:n_spikes-skip_last_n]:
                            val += abs(spike[k] - mean)
                        val /= n_spikes - skip_last_n
                        inc = val / stdev
                        scores.append(inc)
                    else:
                        resp = cand.scoring[k]["miss"]
                        if resp == "const":
                            miss = float(cand.scoring[k][resp])
                        elif resp == "mean_mult":
                            miss = mean * float(cand.scoring[k][resp])
                        else:
                            assert False
                        miss = float(miss)
                        scores.append(miss)
#                        print("Missed %s, penalty = %g" % (k, miss))
                else:
                    assert False
                if abs(sum(scores)) > 1e10:
                    print(k)
                    print(self.summary.scoring)
                    print(self.summary.mean)
                    print(self.summary.stdev)
                    print(cand.summary.scoring)
                    print(cand.summary.mean)
                    print(cand.summary.stdev)
                    assert False
        return scores

    # create summary of feature instances
    #   'summary' is an empty feature object. this must be the same
    #     class as the other feature objects that are being summarized
    def summarize(self, summary):
        if len(self.feature_list) == 0:
            print("Error -- no features were extracted. Summary impossible")
            sys.exit()
        # make dummy dict to verify that all feature instances have
        #   identical features
        # only copy features that are in the glossary. some are for
        #   internal use (eg, t_idx -- time index) and aren't important
        #   here
        superset = {}
        for i in range(len(self.feature_list)):
            fx = self.feature_list[i]
            for k in fx.mean.keys():
                if k in fx.glossary:
                    superset[k] = k
        err = 0
        for k in superset.keys():
            for i in range(len(self.feature_list)):
                fx = self.feature_list[i].mean
                if k not in fx:
                    print("Error - feature '%s' not in all data sets" % k)
                    err += 1
        if err > 0:
            return None
        # all features must be of the same type
        # to ensure this, make programmer specify the type being summarized
        self.summary = summary
        # now set summary means to zero
        for k in fx.keys():
            self.summary.mean[k] = 0
        # now calculate the average of all features
        for i in range(len(self.feature_list)):
            fx = self.feature_list[i]
            for k in fx.mean.keys():
                if k in fx.glossary and fx.mean[k] is not None:
                    self.summary.mean[k] += fx.mean[k]
                    self.summary.stdev[k] = 0.0
        # divide out n to get actual mean
        for k in fx.mean.keys():
            self.summary.mean[k] /= 1.0 * len(self.feature_list)
        # calculate standard deviation
        for i in range(len(self.feature_list)):
            fx = self.feature_list[i]
            for k in fx.mean.keys():
                if k in fx.glossary and fx.mean[k] is not None:
                    mean = self.summary.mean[k]
                    dif = mean - fx.mean[k]
                    self.summary.stdev[k] += dif * dif
        # divide out n and take sqrt to get actual stdev
        fx = self.feature_list[0]
        for k in fx.mean.keys():
            if k in fx.glossary and fx.mean[k] is not None:
                val = self.summary.stdev[k]
                val /= 1.0 * len(self.feature_list)
                self.summary.stdev[k] = math.sqrt(val)
        return self




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.collections import PatchCollection
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import scipy.interpolate as si
from scipy.stats import gaussian_kde
import matplotlib.colorbar as cbar
from mpl_toolkits.axes_grid1 import ImageGrid

import allensdk.brain_observatory.circle_plots as cplots
from contextlib import contextmanager

import numpy as np

SI_RANGE = [ 0, 1.5 ]
P_VALUE_MAX = 0.05
PEAK_DFF_MIN = 3
N_HIST_BINS = 50
STIM_COLOR = "#ccccdd"
STIMULUS_COLOR_MAP = LinearSegmentedColormap.from_list('default',[ [1.0,1.0,1.0,0.0], [.6,.6,.85,1.0] ])
PUPIL_COLOR_MAP = LinearSegmentedColormap.from_list(
    'custom_plasma', [[0.050383, 0.029803, 0.527975],
                      [0.417642, 0.000564, 0.658390],
                      [0.692840, 0.165141, 0.564522],
                      [0.881443, 0.392529, 0.383229],
                      [0.988260, 0.652325, 0.211364],
                      [0.940015, 0.975158, 0.131326]])
EVOKED_COLOR = "#b30000"
SPONTANEOUS_COLOR = "#0000b3"

def plot_cell_correlation(sig_corrs, labels, colors, scale=15):
    if len(sig_corrs) > 1:
        alpha = 1.0 / (len(sig_corrs) + 1)
    else:
        alpha = 1.0

    ax = plt.gca()
    ps = []
    for sig_corr, color, label in zip(sig_corrs, colors, labels):
        ax.hist(sig_corr, bins=30, range=[-1,1],
                histtype='stepfilled',
                facecolor=(.6,.6,.6,alpha), 
                edgecolor=color,
                linewidth=1.5,
                label=label)
                  
    ax.set_xlabel("signal correlation")
    ax.set_ylabel("cell count")
    ax.xaxis.grid(True)
    
    leg = ax.legend(loc='upper left', frameon=False)
    for i, t in enumerate(leg.get_texts()):
        t.set_color(colors[i])
        
    plt.text(.125, .5, u'\u2014', transform=ax.transAxes, 
              horizontalalignment='center', verticalalignment='center',
              weight='bold', size='xx-large')
    plt.text(.875, .5, '+', transform=ax.transAxes, 
              horizontalalignment='center', verticalalignment='center',
              weight='bold', size='xx-large')

def population_correlation_scatter(sig_corrs, noise_corrs, labels, colors, scale=15):
    alpha = max(0.85 - 0.15 * (len(sig_corrs)-1), 0.2)
    ax = plt.gca()
    for sig_corr, noise_corr, color, label in zip(sig_corrs, noise_corrs, colors, labels):
        inds = np.tril_indices(len(sig_corr))
        ax.scatter(sig_corr[inds], noise_corr[inds], 
                   s=scale,
                   color=color, 
                   linewidth=0.5, edgecolor='#333333', 
                   label=label, 
                   alpha=alpha)
    ax.set_xlabel("signal correlation")
    ax.set_ylabel("noise correlation")
    ax.set_xlim([-1,1])
    ax.set_ylim([-1,1])
    leg = ax.legend(loc='upper left', frameon=False)
    for i, t in enumerate(leg.get_texts()):
        t.set_color(colors[i])


def plot_mask_outline(mask, ax, color='k'):
    pim = np.pad(mask, 1, 'constant', constant_values=(0,0))
    hedges = np.argwhere(np.diff(pim, axis=0))
    vedges = np.argwhere(np.diff(pim, axis=1))
    hlines = [ [ [r-.5, c-1.5], [r-.5, c-.5] ] for r,c in hedges ]
    vlines = [ [ [r-1.5, c-.5], [r-.5, c-.5] ] for r,c in vedges ]
    
    for p1,p2 in hlines + vlines:
        ax.add_line(mlines.Line2D([ p1[1], p2[1] ], 
                                  [ p1[0], p2[0] ], 
                                  linewidth=3, 
                                  color=color, 
                                  clip_on=False))
        

class DimensionPatchHandler(object):
    def __init__(self, vals, start_color, end_color, *args, **kwargs):
        super(DimensionPatchHandler, self).__init__(*args, **kwargs)
        self.vals = vals
        self.start_color = start_color
        self.end_color = end_color

    def legend_artist(self, legend, orig_handle, fontsize, handlebox):
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width, height = handlebox.width, handlebox.height

        num_vals = len(self.vals)
        sub_width = float(width) / num_vals
        x = x0
        for i in range(len(self.vals)):
            rgb = self.dim_color(i)
            r = mpatches.Rectangle((x+i*sub_width, y0), 
                                   sub_width, y0+height, 
                                   facecolor=rgb, linewidth=0)

            r.set_clip_on(False)
            handlebox.add_artist(r)
        return r

    def dim_color(self, index):
        rgb1 = np.array(mcolors.colorConverter.to_rgb(self.start_color))
        rgb2 = np.array(mcolors.colorConverter.to_rgb(self.end_color))
        t = float(index) / (len(self.vals)+1)
        rgb = t * rgb2 + (1.0 - t) * rgb1
        return rgb

def float_label(n):
    if isinstance(n, int):
        return str(n)
    if n.is_integer():
        return str(int(n))
    else:
        return "%.2f" % n

def plot_representational_similarity(rs, dims=None, dim_labels=None, colors=None, dim_order=None, labels=True):
    if dim_order is not None:
        rsr = np.arange(len(rs)).reshape(*map(len,dims))
        rsrt = rsr.transpose(dim_order)
        ri = rsrt.flatten()
        rs = rs[ri,:][:,ri]

        dims = np.array(dims)[dim_order]
        colors = np.array(colors)[dim_order]
        dim_labels = np.array(dim_labels)[dim_order]
    
    # force the color map to be centered at zero
    clim = np.nanpercentile(rs, [5.0,95.0], axis=None)
    vrange = max(abs(clim[0]), abs(clim[1]))

    rs = rs.copy()
    np.fill_diagonal(rs, np.nan)
 
    if labels:
        grid = ImageGrid(plt.gcf(), 111,
                         nrows_ncols=(1,1),
                         cbar_location="right",
                         cbar_mode="single",
                         cbar_size="7%",
                         cbar_pad=0.05)
        
        for ax in grid: pass
    else:
        ax = plt.gca()

    im = ax.imshow(rs, interpolation='nearest', cmap='RdBu_r', vmin=-vrange, vmax=vrange)
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.set_xticks([])
    ax.set_yticks([])

    if labels:
        cbar = ax.cax.colorbar(im)
        cbar.set_label_text('stimulus correlation')
    
    if dims is not None:
        dim_labels = ["%s(%s)" % (dim_labels[i],', '.join(map(float_label, dims[i].tolist()))) for i in range(len(dims)) ]
        dim_handlers = [ DimensionPatchHandler(dims[i], colors[i], 'w') for i in range(len(dims)) ]

        n = len(rs)
        for cell_i in range(n):
            idx = np.unravel_index(cell_i, map(len, dims))

            start = -(len(dims))*2
            width = 1.8
            for dim_i, color in enumerate(colors):
                v_i = idx[dim_i]
                rgb = dim_handlers[dim_i].dim_color(v_i)
                r = mpatches.Rectangle((start + dim_i * width, cell_i-.5), 
                                       width, 1.2, 
                                       facecolor=rgb, linewidth=0)
                r.set_clip_on(False)
                ax.add_patch(r)

                r = mpatches.Rectangle((cell_i-.5, start + dim_i * width), 
                                       1.2, width,
                                       facecolor=rgb, linewidth=0)
                r.set_clip_on(False)
                ax.add_patch(r)

        if labels:
            patches = [ mpatches.Patch(label=dim_labels[i]) for i in range(len(dims)) ]
            ax.legend(handles=patches, 
                      handler_map=dict(zip(patches,dim_handlers)),
                      loc='upper left',
                      bbox_to_anchor=(0,0),
                      ncol=2,
                      fontsize=9,
                      frameon=False)

    if labels:
        plt.subplots_adjust(left=0.07,
                            right=.88,
                            wspace=0.0, hspace=0.0)

def plot_condition_histogram(vals, bins, color=STIM_COLOR):
    plt.grid()
    if len(vals) > 0:
        n, hbins, patches = plt.hist(vals,
                                     bins=np.arange(len(bins)+1)+1,
                                     align='left',
                                     normed=False,
                                     rwidth=.8,
                                     color=color,
                                     zorder=3)
    else:
        hbins = np.arange(len(bins)+1)+1
    plt.xticks(hbins[:-1], np.round(bins, 2))
   

def plot_selectivity_cumulative_histogram(sis, 
                                          xlabel,
                                          si_range=SI_RANGE, 
                                          n_hist_bins=N_HIST_BINS, 
                                          color=STIM_COLOR):
    bins = np.linspace(si_range[0], si_range[1], n_hist_bins)
    yticks = np.linspace(0,1,5)
    xticks = np.linspace(si_range[0], si_range[1], 4)
         
    yscale = 1.0
    # this is for normalizing to total # cells, not just significant cells
    # yscale = float(num_cells) / len(osis)

    # orientation selectivity cumulative histogram
    if len(sis) > 0:
        n, bins, patches = plt.hist(sis, normed=True, bins=bins,
                                    cumulative=True, histtype='stepfilled',
                                    color=color)
    plt.xlim(si_range)
    plt.ylim([0,yscale])
    plt.yticks(yticks*yscale, yticks)
    plt.xticks(xticks)
    
    plt.xlabel(xlabel)
    plt.ylabel("fraction of cells")
    plt.grid()

def plot_radial_histogram(angles,
                          counts,
                          all_angles=None,
                          include_labels=False,
                          offset=180.0,
                          direction=-1,
                          closed=False,
                          color=STIM_COLOR):
    if all_angles is None:
        if len(angles) < 2:
            all_angles = np.linspace(0, 315, 8)
        else:
            all_angles = angles

    dth = (all_angles[1] - all_angles[0]) * 0.5

    if len(counts) == 0:
        max_count = 1
    else:
        max_count = max(counts)

    wedges = []
    for count, angle in zip(counts, angles):
        angle = angle*direction + offset
        wedge = mpatches.Wedge((0,0), count, angle-dth, angle+dth)
        wedges.append(wedge)

    wedge_coll = PatchCollection(wedges)
    wedge_coll.set_facecolor(color)
    wedge_coll.set_zorder(2)

    angles_rad = (all_angles*direction + offset)*np.pi/180.0

    if closed:
        border_coll = cplots.radial_circles([max_count])
    else:
        border_coll = cplots.radial_arcs([max_count], 
                                         min(angles_rad), 
                                         max(angles_rad))
    border_coll.set_facecolor((0,0,0,0))
    border_coll.set_zorder(1)

    line_coll = cplots.angle_lines(angles_rad, 0, max_count)
    line_coll.set_edgecolor((0,0,0,1))
    line_coll.set_linestyle(":")
    line_coll.set_zorder(1)

    ax = plt.gca()
    ax.add_collection(wedge_coll)
    ax.add_collection(border_coll)
    ax.add_collection(line_coll)

    if include_labels:
        cplots.add_angle_labels(ax, angles_rad, all_angles.astype(int), max_count, (0,0,0,1), offset=max_count*0.1)
        ax.set(xlim=(-max_count*1.2, max_count*1.2),
               ylim=(-max_count*1.2, max_count*1.2),
               aspect=1.0)
    else:
        ax.set(xlim=(-max_count*1.05, max_count*1.05),
               ylim=(-max_count*1.05, max_count*1.05),
               aspect=1.0)
        
def plot_time_to_peak(msrs, ttps, t_start, t_end, stim_start, stim_end, cmap):
    plt.plot(ttps, np.arange(msrs.shape[0],0,-1)-0.5, color='black')
    if msrs.shape[0] > 0:
        plt.imshow(msrs,
                cmap=cmap, clim=[0,3],
                aspect=float((t_end-t_start) / msrs.shape[0]),  # float to get rid of MPL error
                extent=[t_start, t_end, 0, msrs.shape[0]], interpolation='nearest')
        plt.ylim([0,msrs.shape[0]])
    else:
        plt.ylim([0, 1])
    plt.xlim([t_start, t_end])

    plt.axvline(stim_start, linestyle=':', color='black')
    plt.axvline(stim_end, linestyle=':', color='black')

    xticks = np.array([ t_start, stim_start, stim_end, t_end ]) 
    plt.xticks(xticks, np.round(xticks - stim_start, 2))
    plt.xlabel("time from stimulus start (s)")

    yticks, _ = plt.yticks()
    plt.ylabel("cell number")

@contextmanager
def figure_in_px(w, h, file_name, dpi=96.0, transparent=False):
    fig = plt.figure(figsize=(w/dpi, h/dpi), dpi=dpi)

    yield fig
  
    plt.savefig(file_name, dpi=dpi, transparent=transparent)
    plt.close()

def finalize_no_axes(pad=0.0):
    plt.axis('off')
    plt.subplots_adjust(left=pad, 
                        right=1.0-pad, 
                        bottom=pad, 
                        top=1.0-pad, 
                        wspace=0.0, hspace=0.0)

def finalize_with_axes(pad=.3):
    plt.tight_layout(pad=pad)

def finalize_no_labels(pad=.3, legend=False):
    ax = plt.gca()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    if not legend and ax.legend_ is not None:
        ax.legend_.remove()
    plt.tight_layout(pad=pad)

def plot_combined_speed(binned_resp_vis, binned_dx_vis, binned_resp_sp, binned_dx_sp,
                        evoked_color, spont_color):
    ax = plt.gca()
    num_bins = max(binned_dx_vis.shape[0], binned_dx_sp.shape[0])
    
    plot_speed(binned_resp_vis, binned_dx_vis, num_bins, evoked_color)
    plot_speed(binned_resp_sp, binned_dx_sp, num_bins, spont_color)
    
    xmin = min(binned_dx_vis[:,0].min(), binned_dx_sp[:,0].min())
    xmax = max(binned_dx_vis[:,0].max(), binned_dx_sp[:,0].max())


    ymin = min(binned_resp_vis[:,0].min(), binned_resp_sp[:,0].min())
    ymax = max(binned_resp_vis[:,0].max(), binned_resp_sp[:,0].max())

    xpadding = (xmax-xmin)*.05
    ypadding = (ymax-ymin)*.20

    ax.set_xlim([xmin - xpadding, xmax + xpadding])
    ax.set_ylim([ymin - ypadding, ymax + ypadding])


def plot_speed(binned_resp, binned_dx, num_bins, color): 
    ax = plt.gca()

    # plot the zero bin as a dot with whiskers
    ax.errorbar([ binned_dx[0,0] ], [ binned_resp[0,0] ], yerr=[ binned_resp[0,1] ], fmt='o', color=color)

    # if there's only one bin, drop out
    if len(binned_dx[:,0]) <= 1:
        return

    f = si.interp1d(binned_dx[:,0], binned_resp[:,0])    
    x = np.linspace(min(binned_dx[:,0]), max(binned_dx[:,0]), num=num_bins, endpoint=True)
    y = f(x)
    
    f_up = si.interp1d(binned_dx[:,0], binned_resp[:,0] + binned_resp[:,1])
    y_up = f_up(x)
    
    f_down = si.interp1d(binned_dx[:,0], binned_resp[:,0] - binned_resp[:,1])
    y_down = f_down(x)
    
    ax.plot(x, y, color=color)    
    ax.fill_between(x, y_down, y_up, facecolor=color, alpha=0.1)


def plot_receptive_field(rf, color_map=None, clim=None, 
                         mask=None, outline_color='#cccccc',
                         scalebar=True):
    if mask is not None:
        rf = np.ma.array(rf, mask=~mask)

    if clim is None:
        clim = np.nanpercentile(rf, [1.0,99.0], axis=None)

    plt.imshow(rf, interpolation='nearest', 
               cmap=color_map, 
               clim=clim,
               origin='bottom')

    if mask is not None:
        plot_mask_outline(mask, plt.gca(), outline_color)

    if scalebar:
        scale_dims = np.array([ 28.0, 16.0 ])
        scale_p = [ 26.8, 14.8 ] 
        text_p = [ scale_p[0]+0.5, scale_p[1]-0.5 ]
        
        
        ax = plt.gca()
        ax.add_patch(mpatches.Rectangle(scale_p / scale_dims, 
                                        1.0/scale_dims[0], 1.0/scale_dims[1], 
                                        facecolor='w',
                                        transform=ax.transAxes,
                                        linewidth=1.0,
                                        edgecolor=outline_color))
        plt.text(text_p[0] / scale_dims[0], text_p[1] / scale_dims[1], "4deg",
                 horizontalalignment='center',
                 verticalalignment='center',
                 transform=ax.transAxes)
        


def plot_pupil_location(xy_deg, s=1, c=None, cmap=PUPIL_COLOR_MAP,
                        edgecolor='', include_labels=True):
    if c is None:
        xy_deg = xy_deg[~np.isnan(xy_deg).any(axis=1)]
        c = gaussian_kde(xy_deg.T)(xy_deg.T)
    plt.scatter(xy_deg[:,0], xy_deg[:,1], s=s, c=c, cmap=cmap,
                edgecolor=edgecolor)
    plt.xlim(-70, 70)
    plt.ylim(-70, 70)

    if include_labels:
        plt.xlabel("azimuth (degrees)")
        plt.ylabel("altitude (degrees)")



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
import os
import numpy
from allensdk.model.biophys_sim.config import Config
from allensdk.model.biophysical.utils import Utils
from allensdk.core.dat_utilities import DatUtilities
from allensdk.api.queries.biophysical_api import BiophysicalApi


@pytest.mark.skipif(True,
                    reason="partial testing")
@pytest.mark.xfail
def test_biophysical():
    neuronal_model_id = 472451419    # get this from the web site

    model_directory = '.'

    bp = BiophysicalApi('http://api.brain-map.org')
    bp.cache_stimulus = False  # don't want to download the large stimulus NWB file
    bp.cache_data(neuronal_model_id, working_directory=model_directory)
    os.system('nrnivmodl modfiles')

    description = Config().load('manifest.json')
    utils = Utils(description)
    h = utils.h

    manifest = description.manifest
    morphology_path = manifest.get_path('MORPHOLOGY')
    utils.generate_morphology(morphology_path.encode('ascii', 'ignore'))
    utils.load_cell_parameters()

    stim = h.IClamp(h.soma[0](0.5))
    stim.amp = 0.18
    stim.delay = 1000.0
    stim.dur = 1000.0

    h.tstop = 3000.0

    vec = utils.record_values()

    h.finitialize()
    h.run()

    output_path = 'output_voltage.dat'

    junction_potential = description.data['fitting'][0]['junction_potential']
    mV = 1.0e-3
    ms = 1.0e-3

    output_data = (numpy.array(vec['v']) - junction_potential) * mV
    output_times = numpy.array(vec['t']) * ms

    DatUtilities.save_voltage(output_path, output_data, output_times)
    
    assert numpy.count_nonzero(output_data) > 0



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
import numpy as np
from allensdk.ephys.ephys_extractor import EphysSweepSetFeatureExtractor, input_resistance
import os
path = os.path.dirname(__file__)


def test_extractor_no_values():
    ext = EphysSweepSetFeatureExtractor()


def test_extractor_wrong_inputs():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    i = np.zeros_like(v)

    with pytest.raises(ValueError):
        ext = EphysSweepSetFeatureExtractor(t, v, i)

    with pytest.raises(ValueError):
        ext = EphysSweepSetFeatureExtractor([t], v, i)

    with pytest.raises(ValueError):
        ext = EphysSweepSetFeatureExtractor([t], [v], i)

    with pytest.raises(ValueError):
        ext = EphysSweepSetFeatureExtractor([t, t], [v], [i])

    with pytest.raises(ValueError):
        ext = EphysSweepSetFeatureExtractor([t, t], [v, v], [i])


def test_extractor_on_sample_data():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]

    ext = EphysSweepSetFeatureExtractor([t], [v])
    ext.process_spikes()
    swp = ext.sweeps()[0]
    spikes = swp.spikes()

    keys = swp.spike_feature_keys()
    swp_keys = swp.sweep_feature_keys()
    result = swp.spike_feature(keys[0])
    result = swp.sweep_feature("first_isi")
    result = ext.sweep_features("first_isi")
    result = ext.spike_feature_averages(keys[0])

    with pytest.raises(KeyError):
        result = swp.spike_feature("nonexistent_key")

    with pytest.raises(KeyError):
        result = swp.sweep_feature("nonexistent_key")


def test_extractor_on_sample_data_with_i():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    i = np.zeros_like(v)

    ext = EphysSweepSetFeatureExtractor([t], [v], [i])
    ext.process_spikes()


def test_extractor_on_zero_voltage():
    t = np.arange(0, 4000) * 5e-5
    v = np.zeros_like(t)
    i = np.zeros_like(t)

    ext = EphysSweepSetFeatureExtractor([t], [v], [i])
    ext.process_spikes()


def test_extractor_on_variable_time_step():
    data = np.loadtxt(os.path.join(path, "data/spike_test_var_dt.txt"))
    t = data[:, 0]
    v = data[:, 1]

    ext = EphysSweepSetFeatureExtractor([t], [v])
    ext.process_spikes()
    expected_thresh_ind = np.array([73, 183, 314, 463, 616, 770])
    sweep = ext.sweeps()[0]
    assert np.allclose(sweep.spike_feature("threshold_index"), expected_thresh_ind)


def test_extractor_with_high_init_dvdt():
    data = np.loadtxt(os.path.join(path, "data/spike_test_high_init_dvdt.txt"))
    t = data[:, 0]
    v = data[:, 1]

    ext = EphysSweepSetFeatureExtractor([t], [v])
    ext.process_spikes()
    expected_thresh_ind = np.array([11222, 16256, 24058])
    sweep = ext.sweeps()[0]
    assert np.allclose(sweep.spike_feature("threshold_index"), expected_thresh_ind)


def test_extractor_input_resistance():
    t = np.arange(0, 1.0, 5e-6)
    v1 = np.ones_like(t) * -5.
    v2 = np.ones_like(t) * -10.
    i1 = np.ones_like(t) * -50.
    i2 = np.ones_like(t) * -100.

    ext = EphysSweepSetFeatureExtractor([t, t], [v1, v2], [i1, i2])
    ri = input_resistance(ext)
    assert np.allclose(ri, 100.)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from allensdk.brain_observatory.locally_sparse_noise import LocallySparseNoise
from allensdk.brain_observatory.stimulus_analysis import StimulusAnalysis
import pytest
from mock import patch, MagicMock
import itertools as it


@pytest.fixture
def dataset():
    dataset = MagicMock(name='dataset')

    timestamps = MagicMock(name='timestamps')
    celltraces = MagicMock(name='celltraces')
    dataset.get_corrected_fluorescence_traces = \
        MagicMock(name='get_corrected_fluorescence_traces',
                  return_value=(timestamps, celltraces))
    dataset.get_roi_ids = MagicMock(name='get_roi_ids')
    dataset.get_cell_specimen_ids = MagicMock(name='get_cell_specimen_ids')
    dff_traces = MagicMock(name="dfftraces")
    dataset.get_dff_traces = MagicMock(name='get_dff_traces',
                                       return_value=(None, dff_traces))
    dxcm = MagicMock(name='dxcm')
    dxtime = MagicMock(name='dxtime')
    dataset.get_running_speed=MagicMock(name='get_running_speed',
                                        return_value=(dxcm, dxtime))

    LSN = MagicMock(name='LSN')
    LSN_mask = MagicMock(name='LSN_mask')
    dataset.get_locally_sparse_noise_stimulus_template = \
        MagicMock(name='get_locally_sparse_noise_stimulus_template',
                  return_value=(LSN, LSN_mask))

    return dataset

def mock_speed_tuning():
    binned_dx_sp = MagicMock(name='binned_dx_sp')
    binned_cells_sp = MagicMock(name='binned_cells_sp')
    binned_dx_vis = MagicMock(name='binned_dx_vis')
    binned_cells_vis = MagicMock(name='binned_cells_vis')
    peak_run = MagicMock(name='peak_run')
    
    return MagicMock(name='get_speed_tuning',
                     return_value=(binned_dx_sp,
                                   binned_cells_sp,
                                   binned_dx_vis,
                                   binned_cells_vis,
                                   peak_run))

def mock_sweep_response():
    sweep_response = MagicMock(name='sweep_response')
    mean_sweep_response = MagicMock(name='mean_sweep_response')
    pval = MagicMock(name='pval')
    
    return MagicMock(name='get_sweep_response',
                     return_value=(sweep_response,
                                   mean_sweep_response,
                                   pval))

@patch.object(StimulusAnalysis,
              'get_sweep_response',
              mock_sweep_response())
@patch.object(LocallySparseNoise,
              'get_receptive_field',
              MagicMock(name='get_receptive_field'))
@pytest.mark.parametrize('stimulus,trigger',
                         it.product(('locally_sparse_noise',
                                     'locally_sparse_noise_4deg',
                                     'locally_sparse_noise_8deg'),
                                    (1,2,3,4,5,6)))
def test_harness(dataset,
                 stimulus,
                 trigger):
    with patch('allensdk.brain_observatory.stimulus_analysis.StimulusAnalysis.get_speed_tuning',
               mock_speed_tuning()) as get_speed_tuning:
        lsn = LocallySparseNoise(dataset, stimulus)

        assert lsn._stim_table is StimulusAnalysis._PRELOAD
        assert lsn._LSN is StimulusAnalysis._PRELOAD
        assert lsn._LSN_mask is StimulusAnalysis._PRELOAD
        assert lsn._sweeplength is StimulusAnalysis._PRELOAD
        assert lsn._interlength is StimulusAnalysis._PRELOAD
        assert lsn._extralength is StimulusAnalysis._PRELOAD
        assert lsn._sweep_response is StimulusAnalysis._PRELOAD
        assert lsn._mean_sweep_response is StimulusAnalysis._PRELOAD
        assert lsn._pval is StimulusAnalysis._PRELOAD
        assert lsn._receptive_field is StimulusAnalysis._PRELOAD
    
        if trigger == 1:
            print(lsn.stim_table)
            print(lsn.sweep_response)
            print(lsn.receptive_field)
        elif trigger == 2:
            print(lsn.LSN)
            print(lsn.mean_sweep_response)
            print(lsn.receptive_field)
        elif trigger == 3:
            print(lsn.LSN_mask)
            print(lsn.pval)
            print(lsn.receptive_field)
        elif trigger == 4:
            print(lsn.sweeplength)
            print(lsn.sweep_response)
            print(lsn.receptive_field)
        elif trigger == 5:
            print(lsn.interlength)
            print(lsn.mean_sweep_response)
            print(lsn.receptive_field)
        elif trigger == 6:
            print(lsn.extralength)
            print(lsn.pval)
            print(lsn.receptive_field)

        assert lsn._stim_table is not StimulusAnalysis._PRELOAD
        assert lsn._LSN is not StimulusAnalysis._PRELOAD
        assert lsn._LSN_mask is not StimulusAnalysis._PRELOAD
        assert lsn._sweeplength is not StimulusAnalysis._PRELOAD
        assert lsn._interlength is not StimulusAnalysis._PRELOAD
        assert lsn._extralength is not StimulusAnalysis._PRELOAD
        assert lsn._sweep_response is not StimulusAnalysis._PRELOAD
        assert lsn._mean_sweep_response is not StimulusAnalysis._PRELOAD
        assert lsn._pval is not StimulusAnalysis._PRELOAD
        assert lsn._receptive_field is not StimulusAnalysis._PRELOAD

        # verify super class members weren't preloaded
        assert not dataset.get_corrected_fluorescence_traces.called
        assert lsn._timestamps is StimulusAnalysis._PRELOAD
        assert lsn._celltraces is StimulusAnalysis._PRELOAD
        assert lsn._numbercells is StimulusAnalysis._PRELOAD

        assert not dataset.get_roi_ids.called
        assert lsn._roi_id is StimulusAnalysis._PRELOAD

        assert not dataset.get_cell_specimen_ids.called
        assert lsn._cell_id is StimulusAnalysis._PRELOAD

        assert not dataset.get_dff_traces.called
        assert lsn._dfftraces is StimulusAnalysis._PRELOAD

        assert lsn._dxcm is StimulusAnalysis._PRELOAD
        assert lsn._dxtime is StimulusAnalysis._PRELOAD



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
#test AllenSDK svg api for download and show
from allensdk.api.queries.svg_api import SvgApi
import pytest
import json
from mock import MagicMock
import StringIO
import os

@pytest.fixture
def svg():
    sa = SvgApi()
    return sa
  
def test_build_query(svg):
    ####download true url
    download = True
    groups = None
    section_image_id = 21889
    returned_url = svg.build_query(section_image_id, groups,download)
    assert (returned_url == "http://api.brain-map.org/api/v2/svg_download/21889")
    
    ####download true with one group url
    download = True
    groups = [1]
    section_image_id = 21889
    returned_url = svg.build_query(section_image_id, groups,download)
    assert (returned_url == "http://api.brain-map.org/api/v2/svg_download/21889?groups=1")
    
    ####download true with groups url
    download = True
    groups = [1,2]
    section_image_id = 21889
    returned_url = svg.build_query(section_image_id, groups,download)
    assert (returned_url == "http://api.brain-map.org/api/v2/svg_download/21889?groups=1,2")
    
    ####download false url
    download = False
    groups = None
    section_image_id = 21889
    returned_url = svg.build_query(section_image_id, groups,download)
    assert (returned_url == "http://api.brain-map.org/api/v2/svg/21889")
    
    ####download false groups exist url
    download = False
    groups = [28]
    section_image_id = 21889
    returned_url = svg.build_query(section_image_id, groups,download)
    assert (returned_url == "http://api.brain-map.org/api/v2/svg/21889?groups=28")

def test_download_svg(svg):
    svg.retrieve_file_over_http = MagicMock(name='retrieve_file_over_http')
    section_image_id = 21889
    groups = None
    file_path = None
    
    svg.download_svg(section_image_id,groups,file_path)
    svg.retrieve_file_over_http.assert_called_with('http://api.brain-map.org/api/v2/svg_download/21889', '21889.svg')
    
                     
def test_get_svg(svg):
    svg.retrieve_xml_over_http = MagicMock(name='retrieve_xml_over_http')

    ####groups None
    section_image_id = 100960033
    groups = None

    svg.get_svg(section_image_id, groups)
    svg.retrieve_xml_over_http.assert_called_with("http://api.brain-map.org/api/v2/svg/100960033")
    
    ####groups in 28
    section_image_id = 100960033
    groups = [28]

    svg.get_svg(section_image_id, groups)
    svg.retrieve_xml_over_http.assert_called_with("http://api.brain-map.org/api/v2/svg/100960033?groups=28")



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from mock import patch, MagicMock
from pkg_resources import resource_filename  # @UnresolvedImport
import numpy as np
from allensdk.core.nwb_data_set import NwbDataSet
import pytest
import os

NWB_FLAVORS = []

if 'TEST_EPHYS_NWB_FILES' in os.environ:
    nwb_list_file = os.environ['TEST_EPHYS_NWB_FILES']
else:
    nwb_list_file = resource_filename(__name__, 'nwb_ephys_files.txt')
with open(nwb_list_file, 'r') as f:
    NWB_FLAVORS = [l.strip() for l in f]


@pytest.fixture(params=NWB_FLAVORS)
def data_set(request):
    nwb_file = request.param
    data_set = NwbDataSet(nwb_file)
    return data_set

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_get_sweep_numbers(data_set):
    sweep_numbers = data_set.get_sweep_numbers()

    assert len(sweep_numbers) > 0


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_get_experiment_sweep_numbers(data_set):
    sweep_numbers = data_set.get_experiment_sweep_numbers()

    assert len(sweep_numbers) > 0


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_get_spike_times(data_set):
    sweep_numbers = data_set.get_experiment_sweep_numbers()

    found_spikes = False

    for n in sweep_numbers:
        spike_times = data_set.get_spike_times(n)
        if len(spike_times > 0):
            found_spikes = True

    assert found_spikes is True

def mock_h5py_file(m=None, data=None):
    if m is None:
        m = MagicMock()

    f = MagicMock()

    if data is None:
        f.__enter__.return_value = f
    else:
        f.__enter__.return_value = data

    m.return_value = f

    return m


@pytest.fixture
def mock_data_set():
    nwb_file = 'fixture.nwb'
    data_set = NwbDataSet(nwb_file)
    return data_set


def test_fill_sweep_responses(mock_data_set):
    data_set = mock_data_set
    DATA_LENGTH = 5

    h5 = {
        'stimulus': {
            'presentation': {
                'Sweep_1': {
                    'aibs_stimulus_amplitude_pa': 15.0,
                    'aibs_stimulus_name': 'Joe',
                    'gain': 1.0,
                    'initial_access_resistance': 0.05,
                    'seal': True
                }
            }
        },
        'epochs': {
            'Sweep_1': {
                'description': 'sweep 1 description',
                'stimulus': {},
                'response': {
                    'count': DATA_LENGTH,
                    'idx_start': 0,
                    'timeseries': {
                        'data': np.ones(DATA_LENGTH) * 1.0
                    }
                }
            }
        }
    }

    with patch('h5py.File', mock_h5py_file(data=h5)):
        data_set.fill_sweep_responses(0.0, [1])

    assert not np.any(h5['epochs']['Sweep_1']['response']['timeseries']['data'])
    assert len(h5['epochs']['Sweep_1']['response']['timeseries']['data']) == \
        DATA_LENGTH


@pytest.mark.xfail
def test_set_spike_times(mock_data_set):
    data_set = mock_data_set
    DATA_LENGTH = 5

    h5 = {
        'analysis': {
            'spike_times': {
                'Sweep_1': {}
            }
        },
        'stimulus': {
            'presentation': {
                'Sweep_1': {
                    'aibs_stimulus_amplitude_pa': 15.0,
                    'aibs_stimulus_name': 'Joe',
                    'gain': 1.0,
                    'initial_access_resistance': 0.05,
                    'seal': True
                }
            }
        },
        'epochs': {
            'Sweep_1': {
                'description': 'sweep 1 description',
                'stimulus': {},
                'response': {
                    'count': DATA_LENGTH,
                    'idx_start': 0,
                    'timeseries': {
                        'data': np.ones(DATA_LENGTH) * 1.0
                    }
                }
            }
        }
    }

    with patch('h5py.File', mock_h5py_file(data=h5)):
        data_set.set_spike_times(1, [0.1, 0.2, 0.3, 0.4, 0.5])

    assert False

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_get_sweep_metadata(data_set):
    sweep_metadata = data_set.get_sweep_metadata(1)

    assert sweep_metadata is not None



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import h5py
import logging
import pandas as pd
import numpy as np
import allensdk.brain_observatory.roi_masks as roi
import itertools
from allensdk.brain_observatory.locally_sparse_noise import LocallySparseNoise
import allensdk.brain_observatory.stimulus_info as si
import dateutil
import re
import os
from pkg_resources import parse_version
from allensdk.brain_observatory.brain_observatory_exceptions import (MissingStimulusException,
                                                                     NoEyeTrackingException)
from allensdk.api.cache import memoize

# Deprecation rerouting:
from allensdk.deprecated import deprecated
from allensdk.brain_observatory.stimulus_info import warp_stimulus_coords as si_warp_stimulus_coords
from allensdk.brain_observatory.stimulus_info import make_display_mask as si_make_display_mask
from allensdk.brain_observatory.stimulus_info import mask_stimulus_template as si_mask_stimulus_template


def get_epoch_mask_list(st, threshold, max_cuts=2):
    '''Convenience function to cut a stim table into multiple epochs

    :param st: input stimtable
    :param threshold: threshold on the max duration of a subepoch
    :param max_cuts: maximum number of allowed epochs to cut into
    :return: epoch_mask_list, a list of indices that define the start and end of sub-epochs
    '''

    if threshold is None:
        raise NotImplementedError('threshold not set for this type of session')

    delta = (st.start.values[1:] - st.end.values[:-1])
    cut_inds = np.where(delta > threshold)[0] + 1

    epoch_mask_list = []

    if len(cut_inds) > max_cuts:
        raise Exception('more than 2 epochs cut')

    for ii in range(len(cut_inds)+1):

        if ii == 0:
            first_ind = st.iloc[0].start
        else:
            first_ind = st.iloc[cut_inds[ii-1]].start

        if ii == len(cut_inds):
            last_ind_inclusive = st.iloc[-1].end
        else:
            last_ind_inclusive = st.iloc[cut_inds[ii]-1].end

        epoch_mask_list.append((first_ind,last_ind_inclusive))

    return epoch_mask_list


class BrainObservatoryNwbDataSet(object):
    PIPELINE_DATASET = 'brain_observatory_pipeline'
    SUPPORTED_PIPELINE_VERSION = "2.0"

    FILE_METADATA_MAPPING = {
        'age': 'general/subject/age',
        'sex': 'general/subject/sex',
        'imaging_depth': 'general/optophysiology/imaging_plane_1/imaging depth',
        'targeted_structure': 'general/optophysiology/imaging_plane_1/location',
        'ophys_experiment_id': 'general/session_id',
        'experiment_container_id': 'general/experiment_container_id',
        'device_string': 'general/devices/2-photon microscope',
        'excitation_lambda': 'general/optophysiology/imaging_plane_1/excitation_lambda',
        'indicator': 'general/optophysiology/imaging_plane_1/indicator',
        'fov': 'general/fov',
        'genotype': 'general/subject/genotype',
        'session_start_time': 'session_start_time',
        'session_type': 'general/session_type',
        'specimen_name': 'general/specimen_name',
        'generated_by': 'general/generated_by'
    }

    STIMULUS_TABLE_TYPES = {
        'abstract_feature_series': [si.DRIFTING_GRATINGS, si.STATIC_GRATINGS],
        'indexed_time_series': [si.NATURAL_SCENES, si.LOCALLY_SPARSE_NOISE,
                                si.LOCALLY_SPARSE_NOISE_4DEG, si.LOCALLY_SPARSE_NOISE_8DEG],
        'repeated_indexed_time_series':[si.NATURAL_MOVIE_ONE, si.NATURAL_MOVIE_TWO, si.NATURAL_MOVIE_THREE]

    }

    # this array was moved before file versioning was in place
    MOTION_CORRECTION_DATASETS = [ "MotionCorrection/2p_image_series/xy_translations",
                                   "MotionCorrection/2p_image_series/xy_translation" ]

    def __init__(self, nwb_file):

        self.nwb_file = nwb_file
        self.pipeline_version = None

        if os.path.exists(self.nwb_file):
            meta = self.get_metadata()
            if meta and 'pipeline_version' in meta:
                pipeline_version_str = meta['pipeline_version']
                self.pipeline_version = parse_version(pipeline_version_str)

                if self.pipeline_version > parse_version(self.SUPPORTED_PIPELINE_VERSION):
                    logging.warning("File %s has a pipeline version newer than the version supported by this class (%s vs %s)."
                                    " Please update your AllenSDK." % (nwb_file, pipeline_version_str, self.SUPPORTED_PIPELINE_VERSION))

        self._stimulus_search = None

    def get_stimulus_epoch_table(self):
        '''Returns a pandas dataframe that summarizes the stimulus epoch duration for each acquisition time index in
        the experiment

        Parameters
        ----------
        None

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        traces: 2D numpy array
            Fluorescence traces for each cell
        '''


        # These are thresholds used by get_epoch_mask_list to set a maximum limit on the delta aqusistion frames to
        #  count as different trials (rows in the stim table).  This helps account for dropped frames, so that they dont
        #  cause the cutting of an entire experiment into too many stimulus epochs.  If these thresholds are too low,
        #  the assert statment in get_epoch_mask_list will halt execution.  In that case, make a bug report!.
        threshold_dict = {si.THREE_SESSION_A:31+7,
                          si.THREE_SESSION_B:15,
                          si.THREE_SESSION_C:7,
                          si.THREE_SESSION_C2:7}

        stimulus_table_dict = {}
        for stimulus in self.list_stimuli():

            stimulus_table_dict[stimulus] = self.get_stimulus_table(stimulus)

            if stimulus == si.SPONTANEOUS_ACTIVITY:
                stimulus_table_dict[stimulus]['frame'] = 0

        interval_list = []
        interval_stimulus_dict = {}
        for stimulus in self.list_stimuli():
            stimulus_interval_list = get_epoch_mask_list(stimulus_table_dict[stimulus], threshold=threshold_dict.get(self.get_session_type(), None))
            for stimulus_interval in stimulus_interval_list:
                interval_stimulus_dict[stimulus_interval] = stimulus
            interval_list += stimulus_interval_list
        interval_list.sort(key=lambda x: x[0])

        stimulus_signature_list = ['gap']
        duration_signature_list = [int(interval_list[0][0])]
        interval_signature_list = [(0,int(interval_list[0][0]))]
        for ii, interval in enumerate(interval_list):
            stimulus_signature_list.append(interval_stimulus_dict[interval])
            duration_signature_list.append(int(interval[1] - interval[0]))
            interval_signature_list.append((int(interval[0]), int(interval[1])))

            if ii != len(interval_list)-1:
                stimulus_signature_list.append('gap')
                duration_signature_list.append((int(interval_list[ii+1][0] - interval_list[ii][1])))
                interval_signature_list.append((int(interval_list[ii][1]), int(interval_list[ii+1][0])))

        stimulus_signature_list.append('gap')
        interval_signature_list.append((int(interval_list[-1][1]), len(self.get_fluorescence_timestamps())))
        duration_signature_list.append(interval_signature_list[-1][1]-interval_signature_list[-1][0])

        interval_df = pd.DataFrame({'stimulus':stimulus_signature_list,
                                    'duration':duration_signature_list,
                                    'interval':interval_signature_list})

        # Gaps are ininformative; remove them:
        interval_df = interval_df[interval_df.stimulus != 'gap']
        interval_df['start'] = [x[0] for x in interval_df['interval'].values]
        interval_df['end'] = [x[1] for x in interval_df['interval'].values]

        interval_df.reset_index(inplace=True, drop=True)
        interval_df.drop(['interval', 'duration'], axis=1, inplace=True)
        return interval_df


    def get_fluorescence_traces(self, cell_specimen_ids=None):
        ''' Returns an array of fluorescence traces for all ROI and
        the timestamps for each datapoint

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then all are returned

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        traces: 2D numpy array
            Fluorescence traces for each cell
        '''
        timestamps = self.get_fluorescence_timestamps()
        with h5py.File(self.nwb_file, 'r') as f:
            ds = f['processing'][self.PIPELINE_DATASET][
                'Fluorescence']['imaging_plane_1']['data']

            if cell_specimen_ids is None:
                cell_traces = ds.value
            else:
                inds = self.get_cell_specimen_indices(cell_specimen_ids)
                cell_traces = ds[inds, :]

        return timestamps, cell_traces

    def get_fluorescence_timestamps(self):
        ''' Returns an array of timestamps in seconds for the fluorescence traces '''

        with h5py.File(self.nwb_file, 'r') as f:
            timestamps = f['processing'][self.PIPELINE_DATASET][
                'Fluorescence']['imaging_plane_1']['timestamps'].value
        return timestamps

    def get_neuropil_traces(self, cell_specimen_ids=None):
        ''' Returns an array of neuropil fluorescence traces for all ROIs
        and the timestamps for each datapoint

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then all are returned

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        traces: 2D numpy array
            Neuropil fluorescence traces for each cell
        '''

        timestamps = self.get_fluorescence_timestamps()

        with h5py.File(self.nwb_file, 'r') as f:
            if self.pipeline_version >= parse_version("2.0"):
                ds = f['processing'][self.PIPELINE_DATASET][
                    'Fluorescence']['imaging_plane_1_neuropil_response']['data']
            else:
                ds = f['processing'][self.PIPELINE_DATASET][
                    'Fluorescence']['imaging_plane_1']['neuropil_traces']

            if cell_specimen_ids is None:
                np_traces = ds.value
            else:
                inds = self.get_cell_specimen_indices(cell_specimen_ids)
                np_traces = ds[inds, :]

        return timestamps, np_traces


    def get_neuropil_r(self, cell_specimen_ids=None):
        ''' Returns a scalar value of r for neuropil correction of flourescence traces

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then results for all are returned

        Returns
        -------
        r: 1D numpy array, len(r)=len(cell_specimen_ids)
            Scalar for neuropil subtraction for each cell
        '''

        with h5py.File(self.nwb_file, 'r') as f:
            if self.pipeline_version >= parse_version("2.0"):
                r_ds = f['processing'][self.PIPELINE_DATASET][
                    'Fluorescence']['imaging_plane_1_neuropil_response']['r']
            else:
                r_ds = f['processing'][self.PIPELINE_DATASET][
                    'Fluorescence']['imaging_plane_1']['r']

            if cell_specimen_ids is None:
                r = r_ds.value
            else:
                inds = self.get_cell_specimen_indices(cell_specimen_ids)
                r = r_ds[inds]

        return r

    def get_demixed_traces(self, cell_specimen_ids=None):
        ''' Returns an array of demixed fluorescence traces for all ROIs
        and the timestamps for each datapoint

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then all are returned

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        traces: 2D numpy array
            Demixed fluorescence traces for each cell
        '''

        timestamps = self.get_fluorescence_timestamps()

        with h5py.File(self.nwb_file, 'r') as f:
            ds = f['processing'][self.PIPELINE_DATASET][
                'Fluorescence']['imaging_plane_1_demixed_signal']['data']
            if cell_specimen_ids is None:
                traces = ds.value
            else:
                inds = self.get_cell_specimen_indices(cell_specimen_ids)
                traces = ds[inds, :]

        return timestamps, traces

    def get_corrected_fluorescence_traces(self, cell_specimen_ids=None):
        ''' Returns an array of demixed and neuropil-corrected fluorescence traces
        for all ROIs and the timestamps for each datapoint

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then all are returned

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        traces: 2D numpy array
            Corrected fluorescence traces for each cell
        '''

        # starting in version 2.0, neuropil correction follows trace demixing
        if self.pipeline_version >= parse_version("2.0"):
            timestamps, cell_traces = self.get_demixed_traces(cell_specimen_ids)
        else:
            timestamps, cell_traces = self.get_fluorescence_traces(cell_specimen_ids)

        r = self.get_neuropil_r(cell_specimen_ids)

        _, neuropil_traces = self.get_neuropil_traces(cell_specimen_ids)

        fc = cell_traces - neuropil_traces * r[:, np.newaxis]

        return timestamps, fc

    def get_cell_specimen_indices(self, cell_specimen_ids=None):
        ''' Given a list of cell specimen ids, return their index based on their order in this file.

        Parameters
        ----------
        cell_specimen_ids: list of cell specimen ids

        '''

        all_cell_specimen_ids = list(self.get_cell_specimen_ids())

        try:
            inds = [list(all_cell_specimen_ids).index(i)
                    for i in cell_specimen_ids]
        except ValueError as e:
            raise ValueError("Cell specimen not found (%s)" % str(e))

        return inds

    def get_dff_traces(self, cell_specimen_ids=None):
        ''' Returns an array of dF/F traces for all ROIs and
        the timestamps for each datapoint

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return data for. If this is None (default)
            then all are returned

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        dF/F: 2D numpy array
            dF/F values for each cell
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            dff_ds = f['processing'][self.PIPELINE_DATASET][
                'DfOverF']['imaging_plane_1']

            timestamps = dff_ds['timestamps'].value

            if cell_specimen_ids is None:
                cell_traces = dff_ds['data'].value
            else:
                inds = self.get_cell_specimen_indices(cell_specimen_ids)
                cell_traces = dff_ds['data'][inds, :]

        return timestamps, cell_traces

    def get_roi_ids(self):
        ''' Returns an array of IDs for all ROIs in the file

        Returns
        -------
        ROI IDs: list
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            roi_id = f['processing'][self.PIPELINE_DATASET][
                'ImageSegmentation']['roi_ids'].value
        return roi_id

    def get_cell_specimen_ids(self):
        ''' Returns an array of cell IDs for all cells in the file

        Returns
        -------
        cell specimen IDs: list
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            cell_id = f['processing'][self.PIPELINE_DATASET][
                'ImageSegmentation']['cell_specimen_ids'].value
        return cell_id

    def get_session_type(self):
        ''' Returns the type of experimental session, presently one of the
        following: three_session_A, three_session_B, three_session_C

        Returns
        -------
        session type: string
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            session_type = f['general/session_type'].value
        return session_type

    def get_max_projection(self):
        '''Returns the maximum projection image for the 2P movie.

        Returns
        -------
        max projection: np.ndarray
        '''

        with h5py.File(self.nwb_file, 'r') as f:
            max_projection = f['processing'][self.PIPELINE_DATASET]['ImageSegmentation'][
                'imaging_plane_1']['reference_images']['maximum_intensity_projection_image']['data'].value
        return max_projection

    def list_stimuli(self):
        ''' Return a list of the stimuli presented in the experiment.

        Returns
        -------
        stimuli: list of strings
        '''

        with h5py.File(self.nwb_file, 'r') as f:
            keys = f["stimulus/presentation/"].keys()
        return [k.replace('_stimulus', '') for k in keys]

    def get_stimulus_table(self, stimulus_name):
        ''' Return a stimulus table given a stimulus name '''
        if stimulus_name in self.STIMULUS_TABLE_TYPES['abstract_feature_series']:
            return _get_abstract_feature_series_stimulus_table(self.nwb_file, stimulus_name + "_stimulus")
        elif stimulus_name in self.STIMULUS_TABLE_TYPES['indexed_time_series']:
            return _get_indexed_time_series_stimulus_table(self.nwb_file, stimulus_name)
        elif stimulus_name in self.STIMULUS_TABLE_TYPES['repeated_indexed_time_series']:
            return _get_repeated_indexed_time_series_stimulus_table(self.nwb_file, stimulus_name)
        elif stimulus_name == 'spontaneous':
            return self.get_spontaneous_activity_stimulus_table()
        elif stimulus_name == 'master':

            epoch_table = self.get_stimulus_epoch_table()

            stimulus_table_dict = {}
            for stimulus in self.list_stimuli():
                stimulus_table_dict[stimulus] = self.get_stimulus_table(stimulus)

            table_list = []
            for stimulus in self.list_stimuli():
                curr_stimtable = stimulus_table_dict[stimulus]

                for _, row in epoch_table[epoch_table['stimulus'] == stimulus].iterrows():

                    epoch_start_ind, epoch_end_ind = row['start'], row['end']
                    curr_subtable = curr_stimtable[(epoch_start_ind <= curr_stimtable['start']) &
                                                   (curr_stimtable['end'] <= epoch_end_ind)].copy()
                    curr_subtable['stimulus'] = stimulus
                    table_list.append(curr_subtable)

            table_list = sorted(table_list, key=lambda t: t.iloc[0]['start'])

            new_table = pd.concat(table_list)
            new_table.reset_index(drop=True, inplace=True)

            return new_table



        else:
            raise IOError(
                "Could not find a stimulus table named '%s'" % stimulus_name)

    def get_spontaneous_activity_stimulus_table(self):
        ''' Return the spontaneous activity stimulus table, if it exists.

        Returns
        -------
        stimulus table: pd.DataFrame
        '''
        k = "stimulus/presentation/spontaneous_stimulus"
        with h5py.File(self.nwb_file, 'r') as f:
            events = f[k + '/data'].value
            frame_dur = f[k + '/frame_duration'].value

        start_inds = np.where(events == 1)
        stop_inds = np.where(events == -1)

        if len(start_inds) != len(stop_inds):
            raise Exception(
                "inconsistent start and time times in spontaneous activity stimulus table")

        stim_data = np.column_stack(
            [frame_dur[start_inds, 0].T, frame_dur[stop_inds, 0].T]).astype(int)

        stimulus_table = pd.DataFrame(stim_data, columns=['start', 'end'])

        return stimulus_table

    @memoize
    def get_stimulus_template(self, stimulus_name):
        ''' Return an array of the stimulus template for the specified stimulus.

        Parameters
        ----------
        stimulus_name: string
            Must be one of the strings returned by list_stimuli().

        Returns
        -------
        stimulus table: pd.DataFrame
        '''
        stim_name = stimulus_name + "_image_stack"
        with h5py.File(self.nwb_file, 'r') as f:
            image_stack = f['stimulus']['templates'][stim_name]['data'].value
        return image_stack

    def get_locally_sparse_noise_stimulus_template(self,
                                                   stimulus,
                                                   mask_off_screen=True):
        ''' Return an array of the stimulus template for the specified stimulus.

        Parameters
        ----------
        stimulus: string
           Which locally sparse noise stimulus to retrieve.  Must be one of:
               stimulus_info.LOCALLY_SPARSE_NOISE
               stimulus_info.LOCALLY_SPARSE_NOISE_4DEG
               stimulus_info.LOCALLY_SPARSE_NOISE_8DEG

        mask_off_screen: boolean
           Set off-screen regions of the stimulus to LocallySparseNoise.LSN_OFF_SCREEN.

        Returns
        -------
        tuple: (template, off-screen mask)
        '''

        if stimulus not in si.LOCALLY_SPARSE_NOISE_DIMENSIONS:
            raise KeyError("%s is not a known locally sparse noise stimulus" % stimulus)

        template = self.get_stimulus_template(stimulus)

        # build mapping from template coordinates to display coordinates
        template_shape = si.LOCALLY_SPARSE_NOISE_DIMENSIONS[stimulus]
        template_shape = [ template_shape[1], template_shape[0] ]

        template_display_shape = (1260, 720)
        display_shape = (1920, 1200)

        scale = [
            float(template_shape[0]) / float(template_display_shape[0]),
            float(template_shape[1]) / float(template_display_shape[1])
        ]
        offset = [
            -(display_shape[0] - template_display_shape[0]) * 0.5,
            -(display_shape[1] - template_display_shape[1]) * 0.5
        ]

        x, y = np.meshgrid(np.arange(display_shape[0]), np.arange(
            display_shape[1]), indexing='ij')
        template_display_coords = np.array([(x + offset[0]) * scale[0] - 0.5,
                                            (y + offset[1]) * scale[1] - 0.5],
                                           dtype=float)
        template_display_coords = np.rint(template_display_coords).astype(int)

        # build mask
        template_mask, template_frac = mask_stimulus_template(
            template_display_coords, template_shape)

        if mask_off_screen:
            template[:, ~template_mask.T] = LocallySparseNoise.LSN_OFF_SCREEN

        return template, template_mask.T

    def get_roi_mask_array(self, cell_specimen_ids=None):
        ''' Return a numpy array containing all of the ROI masks for requested cells.
        If cell_specimen_ids is omitted, return all masks.

        Parameters
        ----------
        cell_specimen_ids: list
            List of cell specimen ids.  Default None.

        Returns
        -------
        np.ndarray: NxWxH array, where N is number of cells
        '''

        roi_masks = self.get_roi_mask(cell_specimen_ids)

        if len(roi_masks) == 0:
            raise IOError("no masks found for given cell specimen ids")

        roi_arr = roi.create_roi_mask_array(roi_masks)

        return roi_arr

    def get_roi_mask(self, cell_specimen_ids=None):
        ''' Returns an array of all the ROI masks

        Parameters
        ----------
        cell specimen IDs: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then all are returned

        Returns
        -------
            List of ROI_Mask objects
        '''

        with h5py.File(self.nwb_file, 'r') as f:
            mask_loc = f['processing'][self.PIPELINE_DATASET][
                'ImageSegmentation']['imaging_plane_1']
            roi_list = f['processing'][self.PIPELINE_DATASET][
                'ImageSegmentation']['imaging_plane_1']['roi_list'].value

            inds = None
            if cell_specimen_ids is None:
                inds = range(self.number_of_cells)
            else:
                inds = self.get_cell_specimen_indices(cell_specimen_ids)

            roi_array = []
            for i in inds:
                v = roi_list[i]
                roi_mask = mask_loc[v]["img_mask"].value
                m = roi.create_roi_mask(roi_mask.shape[1], roi_mask.shape[0],
                                        [0, 0, 0, 0], roi_mask=roi_mask, label=v)
                roi_array.append(m)

        return roi_array

    @property
    def number_of_cells(self):
        '''Number of cells in the experiment'''

        # Replace here is there is a better way to get this info:
        return len(self.get_cell_specimen_ids())


    def get_metadata(self):
        ''' Returns a dictionary of meta data associated with each
        experiment, including Cre line, specimen number,
        visual area imaged, imaging depth

        Returns
        -------
        metadata: dictionary
        '''

        meta = {}

        with h5py.File(self.nwb_file, 'r') as f:
            for memory_key, disk_key in BrainObservatoryNwbDataSet.FILE_METADATA_MAPPING.items():
                try:
                    v = f[disk_key].value

                    # convert numpy strings to python strings
                    if v.dtype.type is np.string_:
                        if len(v.shape) == 0:
                            v = v.decode('UTF-8')
                        elif len(v.shape) == 1:
                            v = [ s.decode('UTF-8') for s in v ]
                        else:
                            raise Exception("Unrecognized metadata formatting for field %s" % disk_key)

                    meta[memory_key] = v
                except KeyError as e:
                    logging.warning("could not find key %s", disk_key)

        # extract cre line from genotype string
        genotype = meta.get('genotype')
        meta['cre_line'] = meta['genotype'].split(';')[0] if genotype else None

        imaging_depth = meta.pop('imaging_depth', None)
        meta['imaging_depth_um'] = int(imaging_depth.split()[0]) if imaging_depth else None

        ophys_experiment_id = meta.get('ophys_experiment_id')
        meta['ophys_experiment_id'] = int(ophys_experiment_id) if ophys_experiment_id else None

        experiment_container_id = meta.get('experiment_container_id')
        meta['experiment_container_id'] = int(experiment_container_id) if experiment_container_id else None

        # convert start time to a date object
        session_start_time = meta.get('session_start_time')
        if isinstance(session_start_time, basestring):
            meta['session_start_time'] = dateutil.parser.parse(session_start_time)

        age = meta.pop('age', None)
        if age:
            # parse the age in days
            m = re.match("(.*?) days", age)
            if m:
                meta['age_days'] = int(m.groups()[0])
            else:
                raise IOError("Could not parse age.")


        # parse the device string (ugly, sorry)
        device_string = meta.pop('device_string', None)
        if device_string:
            m = re.match("(.*?)\.\s(.*?)\sPlease*", device_string)
            if m:
                device, device_name = m.groups()
                meta['device'] = device
                meta['device_name'] = device_name
            else:
                raise IOError("Could not parse device string.")

        # file version
        generated_by = meta.pop('generated_by', None)
        version = generated_by[-1] if generated_by else "0.9"
        meta["pipeline_version"] = version

        return meta

    def get_running_speed(self):
        ''' Returns the mouse running speed in cm/s
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            dx_ds = f['processing'][self.PIPELINE_DATASET][
                'BehavioralTimeSeries']['running_speed']
            dxcm = dx_ds['data'].value
            dxtime = dx_ds['timestamps'].value

        timestamps = self.get_fluorescence_timestamps()

        # v0.9 stored this as an Nx1 array instead of a flat 1-d array
        if len(dxcm.shape) == 2:
            dxcm = dxcm[:, 0]

        dxcm, dxtime = align_running_speed(dxcm, dxtime, timestamps)

        return dxcm, dxtime

    def get_pupil_location(self, as_spherical=True):
        '''Returns the x, y pupil location.

        Parameters
        ----------
        as_spherical : bool
            Whether to return the location as spherical (default) or
            not. If true, the result is altitude and azimuth in
            degrees, otherwise it is x, y in centimeters. (0,0) is
            the center of the monitor.

        Returns
        -------
        (timestamps, location)
            Timestamps is an (Nx1) array of timestamps in seconds.
            Location is an (Nx2) array of spatial location.
        '''
        if as_spherical:
            location_key = "pupil_location_spherical"
        else:
            location_key = "pupil_location"
        try:
            with h5py.File(self.nwb_file, 'r') as f:
                eye_tracking = f['processing'][self.PIPELINE_DATASET][
                    'EyeTracking'][location_key]
                pupil_location = eye_tracking['data'].value
                pupil_times = eye_tracking['timestamps'].value
        except KeyError:
            raise NoEyeTrackingException("No eye tracking for this experiment.")

        return pupil_times, pupil_location

    def get_pupil_size(self):
        '''Returns the pupil area in pixels.

        Returns
        -------
        (timestamps, areas)
            Timestamps is an (Nx1) array of timestamps in seconds.
            Areas is an (Nx1) array of pupil areas in pixels.
        '''
        try:
            with h5py.File(self.nwb_file, 'r') as f:
                pupil_tracking = f['processing'][self.PIPELINE_DATASET][
                    'PupilTracking']['pupil_size']
                pupil_size = pupil_tracking['data'].value
                pupil_times = pupil_tracking['timestamps'].value
        except KeyError:
            raise NoEyeTrackingException("No pupil tracking for this experiment.")

        return pupil_times, pupil_size

    def get_motion_correction(self):
        ''' Returns a Panda DataFrame containing the x- and y- translation of each image used for image alignment
        '''

        motion_correction = None
        with h5py.File(self.nwb_file, 'r') as f:
            pipeline_ds = f['processing'][self.PIPELINE_DATASET]

            # pipeline 0.9 stores this in xy_translations
            # pipeline 1.0 stores this in xy_translation
            for mc_ds_name in self.MOTION_CORRECTION_DATASETS:
                try:
                    mc_ds = pipeline_ds[mc_ds_name]

                    motion_log = mc_ds['data'].value
                    motion_time = mc_ds['timestamps'].value
                    motion_names = mc_ds['feature_description'].value

                    motion_correction = pd.DataFrame(motion_log, columns=motion_names)
                    motion_correction['timestamp'] = motion_time

                    # break out if we found it
                    break
                except KeyError as e:
                    pass

        if motion_correction is None:
            raise KeyError("Could not find motion correction data.")

        return motion_correction

    def save_analysis_dataframes(self, *tables):
        store = pd.HDFStore(self.nwb_file, mode='a')
        for k, v in tables:
            store.put('analysis/%s' % (k), v)
        store.close()

    def save_analysis_arrays(self, *datasets):
        with h5py.File(self.nwb_file, 'a') as f:
            for k, v in datasets:
                if k in f['analysis']:
                    del f['analysis'][k]
                f.create_dataset('analysis/%s' % k, data=v)

    @property
    def stimulus_search(self):

        if self._stimulus_search is None:
            self._stimulus_search = si.StimulusSearch(self)
        return self._stimulus_search

    def get_stimulus(self, frame_ind):

        search_result = self.stimulus_search.search(frame_ind)

        if search_result is None or search_result[2]['stimulus'] == si.SPONTANEOUS_ACTIVITY:
            return None, None

        else:

            curr_stimulus = search_result[2]['stimulus']
            if curr_stimulus in si.LOCALLY_SPARSE_NOISE_STIMULUS_TYPES + si.NATURAL_MOVIE_STIMULUS_TYPES + [si.NATURAL_SCENES]:
                curr_frame = search_result[2]['frame']
                return search_result, self.get_stimulus_template(curr_stimulus)[int(curr_frame), :, :]
            elif curr_stimulus == si.STATIC_GRATINGS or curr_stimulus == si.DRIFTING_GRATINGS:
                return search_result, None

def align_running_speed(dxcm, dxtime, timestamps):
    ''' If running speed timestamps differ from fluorescence
    timestamps, adjust by inserting NaNs to running speed.

    Returns
    -------
    tuple: dxcm, dxtime
    '''
    if dxtime[0] != timestamps[0]:
        adjust = np.where(timestamps == dxtime[0])[0][0]
        dxtime = np.insert(dxtime, 0, timestamps[:adjust])
        dxcm = np.insert(dxcm, 0, np.repeat(np.NaN, adjust))
    adjust = len(timestamps) - len(dxtime)
    if adjust > 0:
        dxtime = np.append(dxtime, timestamps[(-1 * adjust):])
        dxcm = np.append(dxcm, np.repeat(np.NaN, adjust))

    return dxcm, dxtime

#
@deprecated('Use allensdk.brain_observatory.stimulus_info.warp_stimulus_coords instead')
def warp_stimulus_coords(*args, **kwargs):
    return si_warp_stimulus_coords(*args, **kwargs)


@deprecated('Use allensdk.brain_observatory.stimulus_info.make_display_mask instead')
def make_display_mask(*args, **kwargs):
    return si_make_display_mask(*args, **kwargs)



@deprecated('Use allensdk.brain_observatory.stimulus_info.mask_stimulus_template instead')
def mask_stimulus_template(*args, **kwargs):
    return si_mask_stimulus_template(*args, **kwargs)


def _get_abstract_feature_series_stimulus_table(nwb_file, stimulus_name):
    ''' Return the a stimulus table for an abstract feature series.

    Returns
    -------
    stimulus table: pd.DataFrame
    '''


    k = "stimulus/presentation/%s" % stimulus_name

    with h5py.File(nwb_file, 'r') as f:
        if k not in f:
            raise MissingStimulusException(
                "Stimulus not found: %s" % stimulus_name)
        stim_data = f[k + '/data'].value
        features = [ v.decode('UTF-8') for v in f[k + '/features'].value ]
        frame_dur = f[k + '/frame_duration'].value

    stimulus_table = pd.DataFrame(stim_data, columns=features)
    stimulus_table.loc[:, 'start'] = frame_dur[:, 0].astype(int)
    stimulus_table.loc[:, 'end'] = frame_dur[:, 1].astype(int)

    return stimulus_table


def _get_indexed_time_series_stimulus_table(nwb_file, stimulus_name):
    ''' Return the a stimulus table for an indexed time series.

    Returns
    -------
    stimulus table: pd.DataFrame
    '''

    k = "stimulus/presentation/%s" % stimulus_name

    with h5py.File(nwb_file, 'r') as f:
        if k not in f:
            k = "stimulus/presentation/%s" % (stimulus_name + "_stimulus")
            if k not in f:
                raise MissingStimulusException("Stimulus not found: %s" % stimulus_name)
        inds = f[k + '/data'].value
        frame_dur = f[k + '/frame_duration'].value

    stimulus_table = pd.DataFrame(inds, columns=['frame'])
    stimulus_table.loc[:, 'start'] = frame_dur[:, 0].astype(int)
    stimulus_table.loc[:, 'end'] = frame_dur[:, 1].astype(int)

    return stimulus_table

def _get_repeated_indexed_time_series_stimulus_table(nwb_file, stimulus_name):

    stimulus_table = _get_indexed_time_series_stimulus_table(nwb_file, stimulus_name)
    a = stimulus_table.groupby(by='frame')

    # If this ever occurs, the repeat counter cant be trusted!
    assert np.floor(len(stimulus_table))/len(a) == int(len(stimulus_table))/len(a)

    stimulus_table['repeat'] = np.repeat(range(len(stimulus_table)/len(a)), len(a))

    return stimulus_table



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
""" The methods in this module are used for configuring dynamics and reset rules for the GlifNeuron.  
For more details on how to use these methods, see :doc:`glif_models`.
"""
import functools
import numpy as np


class GlifNeuronMethod( object ):
    """ A simple class to keep track of the name and parameters associated with a neuron method.
    This class is initialized with a name, function, and parameters to pass to the function.  The
    function then has those passed parameters fixed to a partial function using functools.partial.  
    This class then mimics a function itself using the __call__ convention. Parameters that are not 
    fixed in this way are assumed to be passed into the method when it is called.  If the passed
    parameters contain an argument that is not part of the function signature, an exception will
    be raised.    

    Parameters
    ----------
    method_name : string
        A shorthand name that will be used to reference this method in the `GlifNeuron`.
    method : function
        A python function to be called when this instance is called.
    method_params : dict
        A dictionary mapping function arguments to values for values that should be fixed.
    """

    def __init__(self, method_name, method, method_params):
        self.name = method_name
        self.params = method_params
        self.method = functools.partial(method, **method_params)

    def __call__(self, *args, **kwargs):
        """ Defining this method allows an instance to be called like a function """
        return self.method(*args, **kwargs)

    def to_dict(self):
        return {
            'name': self.name,
            'params': self.params
        }

    def modify_parameter(self, param, operator):
        """ Modify a function parameter needs to be modified after initialization. 

        Parameters
        ----------
        param : string
            the name of the parameter to modify
        operator : callable
            a function or lambda that returns the desired modified value

        Returns
        -------
        type
            the new value of the variable that was just modified.
        """
        value = operator(self.method.keywords[param])
        self.method.keywords[param] = value
        return value


def max_of_line_and_const(x,b,c,d):
    #TODO: move to other library
    """ Find the maximum of a value and a position on a line 

    Parameters
    ----------
    x: float
        x position on line 1
    c: float
        slope of line 1
    d: float
        y-intercept of line 1
    b: float
        y-intercept of line 2

    Returns
    -------
    float
        the max of a line value and a constant
    """

    one = b
    two = c*x+d
    return np.maximum(one,two)


def min_of_line_and_zero(x,c,d):
    #TODO: move to other library
    """ Find the minimum of a value and a position on a line 

    Parameters
    ----------
    x: float
        x position on line 1
    c: float
        slope of line 1
    d: float
        y-intercept of line 1
    b: float
        y-intercept of line 2

    Returns
    -------
    float
        the max of a line value and a constant
    """

    one = 0
    two = c*x+d
    return np.minimum(one,two)


def dynamics_AScurrent_exp(neuron, AScurrents_t0, time_step, spike_time_steps):
    """ Exponential afterspike current dynamics method takes a current at t0 and returns the current at
    a time step later.
    """

    return AScurrents_t0*np.exp(-neuron.k*neuron.dt)
        

def dynamics_AScurrent_none(neuron, AScurrents_t0, time_step, spike_time_steps):
    """ This method always returns zeros for the afterspike currents, regardless of input. """
    return np.zeros(len(AScurrents_t0))


def dynamics_voltage_linear_forward_euler(neuron, voltage_t0, AScurrents_t0, inj):
    """ (TODO) Linear voltage dynamics. """
    return voltage_t0 + (inj + np.sum(AScurrents_t0) - neuron.G * neuron.coeffs['G'] * (voltage_t0 - neuron.El)) * neuron.dt / (neuron.C * neuron.coeffs['C'])

def dynamics_voltage_linear_exact(neuron, voltage_t0, AScurrents_t0, inj):
    """ (TODO) Linear voltage dynamics. """

    C = (neuron.C * neuron.coeffs['C'])
    I = inj + np.sum(AScurrents_t0)
    g = neuron.G * neuron.coeffs['G']
    tau = g/C
    N = (I+ g*neuron.El)/C

    return voltage_t0*np.exp(-neuron.dt*tau) + N*(1-np.exp(-tau*neuron.dt))/tau

def spike_component_of_threshold_forward_euler(th_t0, b_spike, dt):
    '''Spike component of threshold modeled as an exponential decay. Implemented 
    here for forward Euler 
    
    Parameters
    ----------
    th_t0 : float
        threshold input to function
    b_spike : float
        decay constant of exponential
    dt : float
        time step
    '''
    b_spike=-b_spike #TODO: this is here because b_spike is always input as positive although it is negative
    return th_t0 + th_t0*b_spike * dt                                                                         
    
def spike_component_of_threshold_exact(th0, b_spike, t):
    '''Spike component of threshold modeled as an exponential decay. Implemented 
    here as exact analytical solution. 
    
    Parameters
    ----------
    th0 : float
        threshold input to function
    b_spike : float
        decay constant of exponential
    t : float or array
        time step if used in an Euler setup
        time if used analytically
    '''
    b_spike=-b_spike
    return th0*np.exp(b_spike * t)

def voltage_component_of_threshold_forward_euler(th_t0, v_t0, dt, a_voltage, b_voltage, El):
    '''Equation 2.1 of Mihalas and Nieber, 2009 implemented for use in forward Euler. Note 
    here all variables are in reference to threshold infinity.  Therefore thr_inf is zero 
    here (replaced threshold_inf with 0 in the equation to be verbose). This is done so that 
    th_inf can be optimized without affecting this function. 
    
    Parameters
    ----------
    th_t0 : float
        threshold input to function
    v_t0 : float
        voltage input to function
    dt : float
        time step
    a_voltage : float
        constant a
    b_voltage : float
        constant b
    El : float
        reversal potential
    '''
    return th_t0 + (a_voltage*(v_t0-El)-b_voltage*(th_t0-0))*dt

def voltage_component_of_threshold_exact(th0, v0, I, t, a_voltage, b_voltage, C, g, El):
    '''Note this function is the exact formulation; however, dt is used because t0 is the initial time and dt
    is the time the function is exactly evaluated at. Note: that here, this equation is in reference to th_inf.
    Therefore th0 is the total threshold-thr_inf (threshold_inf replaced with 0 in the equation to be verbose).  
    This is done so that th_inf can be optimized without affecting this function.
    
    Parameters
    ----------
    th0 : float
        threshold input to function
    v0 : float
        voltage input to function
    I : float
        total current entering neuron (note if there are after spike currents these must be included in this value)
    t : float or array
        time step if used in an Euler setup
        time if used analytically
    a_voltage : float
        constant a
    b_voltage : float
        constant b
    C : float
        capacitance
    g : float
        conductance (1/resistance)
    El : float 
        reversal potential
    '''
    beta=(I+g*El)/g
    phi=a_voltage/(b_voltage-g/C)
    return phi*(v0-beta)*np.exp(-g*t/C)+1/(np.exp(b_voltage*t))*(th0-phi*(v0-beta)-
                        (a_voltage/b_voltage)*(beta-El)-0) +(a_voltage/b_voltage)*(beta-El) +0


def dynamics_threshold_three_components_exact(neuron, threshold_t0, voltage_t0, AScurrents_t0, inj,
                                              a_spike, b_spike, a_voltage, b_voltage):
    """Analytical solution for threshold dynamics. The threshold will adapt via two mechanisms: 
    1. a voltage dependent adaptation.  
    2. a component initiated by a spike which decays as an exponential.  
    These two component are in reference to threshold infinity and are recorded 
    in the neuron's threshold components.
    The third component refers to th_inf which is added separately as opposed to being 
    included in the voltage component of the threshold as is done in equation 2.1 of
    Mihalas and Nieber 2009.  Threshold infinity is removed for simple optimization.
    
    Parameters
    ----------
    neuron : class
    threshold_t0 : float
        threshold input to function
    voltage_t0 : float
        voltage input to function
    AScurrents_t0 : vector
        values of after spike currents 
    inj : float
        current injected into the neuron
    """
    #TODO: just having the get_threshold_components added an erroneous zero to the beginning of the list 
    if neuron.threshold_components is None:
        neuron.threshold_components = { 'spike': [], 'voltage': [] }
        th_spike = 0
        th_voltage = 0
    else:                
        tcs = neuron.threshold_components
        th_spike = tcs['spike'][-1]
        th_voltage = tcs['voltage'][-1] 

    a_voltage = a_voltage * neuron.coeffs['a']
    b_voltage = b_voltage * neuron.coeffs['b']

    I = inj + np.sum(AScurrents_t0)
    C = neuron.C * neuron.coeffs['C']
    g = neuron.G * neuron.coeffs['G']

    voltage_component=voltage_component_of_threshold_exact(th_voltage, voltage_t0, I, neuron.dt, a_voltage, b_voltage, C, g, neuron.El)
    spike_component = spike_component_of_threshold_exact(th_spike, b_spike, neuron.dt)
 
    #------update the voltage and spiking values of the the
    neuron.append_threshold_components(spike_component, voltage_component)
     
    return voltage_component+spike_component+neuron.th_inf * neuron.coeffs['th_inf']    

def dynamics_threshold_spike_component(neuron, threshold_t0, voltage_t0, AScurrents_t0, inj,
                                       a_spike, b_spike, a_voltage, b_voltage):
    """Analytical solution for spike component of threshold. The threshold will adapt via 
    a component initiated by a spike which decays as an exponential.  The component is in 
    reference to threshold infinity and are recorded in the neuron's threshold components.  The voltage 
    component of the threshold is set to zero in the threshold components  because it is zero here  
    The third component refers to th_inf which is added separately as opposed to being 
    included in the voltage component of the threshold as is done in equation 2.1 of
    Mihalas and Nieber 2009.  Threshold infinity is removed for simple optimization.
    
    Parameters
    ----------
    neuron : class
    threshold_t0 : float
        threshold input to function
    voltage_t0 : float
        voltage input to function
    AScurrents_t0 : vector
        values of after spike currents 
    inj : float
        current injected into the neuron
    """

    #TODO: just having the get_threshold_components added an erroneous zero to the beginning of the list 
    if neuron.threshold_components is None:
        neuron.threshold_components = { 'spike': [], 'voltage': [] }
        th_spike = 0
        th_voltage = 0
    else:                
        tcs = neuron.threshold_components
        th_spike = tcs['spike'][-1]
        th_voltage = tcs['voltage'][-1] 

    spike_component = spike_component_of_threshold_exact(th_spike, b_spike, neuron.dt)
 
    #------update the voltage and spiking values of the the
    neuron.append_threshold_components(spike_component, 0.0)
     
    return spike_component+neuron.th_inf * neuron.coeffs['th_inf']


def dynamics_threshold_inf(neuron, threshold_t0, voltage_t0, AScurrents_t0, inj):
    """ Set threshold to the neuron's instantaneous threshold. 

    Parameters
    ----------
    neuron : class
    threshold_t0 : not used here
    voltage_t0 : not used here
    AScurrents_t0 : not used here 
    inj : not used here
    AScurrents_t0 : not used here 
    inj : not used here
    """   
    return neuron.coeffs['th_inf'] * neuron.th_inf


def reset_AScurrent_sum(neuron, AScurrents_t0, r):
    """ Reset afterspike currents by adding summed exponentials. Left over currents from last spikes as 
    well as newly initiated currents from current spike.  Currents amplitudes in neuron.asc_amp_array need
    to be the amplitudes advanced though the spike cutting.  I.e. In the preprocessor if the after spike currents 
    are calculated via the GLM from spike initiation the amplitude at the time after the spike cutting needs to 
    be calculated and neuron.asc_amp_array needs to be set to this value.
    
    Parameters
    ----------
    r : np.ndarray
        a coefficient vector applied to the afterspike currents
    """
    new_currents=neuron.asc_amp_array * neuron.coeffs['asc_amp_array'] #neuron.asc_amp_array are amplitudes initiating after the spike is cut
    left_over_currents=AScurrents_t0 * r * np.exp(-(neuron.k * neuron.dt * neuron.spike_cut_length)) #advancing cut currents though the spike    

    return new_currents+left_over_currents


def reset_AScurrent_none(neuron, AScurrents_t0):
    """ Reset afterspike currents to zero. """
    
    if np.sum(AScurrents_t0)!=0:
        raise Exception('You are running a LIF but the AScurrents are not zero!')
    return np.zeros(len(AScurrents_t0))


def reset_voltage_v_before(neuron, voltage_t0, a, b):
    """ Reset voltage to the previous value with a scale and offset applied.

    Parameters
    ----------
    a : float
        voltage scale constant
    b : float
        voltage offset constant
    """

    return a*(voltage_t0)+b

def reset_voltage_zero(neuron, voltage_t0):
    """ Reset voltage to zero. """
    return 0.0

def reset_threshold_inf(neuron, threshold_t0, voltage_v1):
    """ Reset the threshold to instantaneous threshold. """
    return neuron.coeffs['th_inf'] * neuron.th_inf

def reset_threshold_three_components(neuron, threshold_t0, voltage_v1, a_spike, b_spike):
    '''This method calculates the two components of the threshold: a spike (fast)
    component and a voltage (slow) component.  The threshold_components vectors are then 
    updated so that the traces match the voltage, current, and total threshold traces.  The
    spike component of the threshold decays via an exponential fit specified by the amplitude
    a_spike and the time constant b_spike fit via the multiblip data.  The voltage component 
    does not change during the duration of the spike.  The 
    spike component are threshold component are summed along with threshold infinity to 
    return the total threshold.  Note that in the current implementation a_spike is added to 
    the last value of the threshold_components which means that a_spike is the amplitude after 
    spike cutting (if there is any).  

    Inputs:  
        neuron: class
            contains attributes of the neuron
        threshold_t0, voltage_t0: float
            are not used but are here for consistency with other methods
        a_spike: float 
            amplitude of the exponential decay of spike component of threshold after spike
            cutting has been implemented.
        b_spike: float
            amplitude of the exponential decay of spike component of threshold
            
    Outputs:
        Returns: float
            the total threshold which is the sum of the spike component of threshold, the voltage 
            component of threshold and threshold infinity (with it's corresponding coefficient)
        neuron.threshold_components: dictionary containing
            a spike: list
                vector of spiking component of threshold that corresponds to the voltage, current, 
                and total threshold traces
            b_spike: list
               vector of voltage component of threshold that corresponds to the voltage, current, 
                and total threshold traces.
                
    Note that this function can be changed to use a_spike at the time of the spike and then have the 
    the spike component plus the residual decay thought the spike.  There are benefits and drawbacks to
    this.  This potential change would be beneficial as it perhaps makes more biological sense for the 
    threshold to go up at the time of spike if the traces are ever used.  Also this would mean that a_spike
    would not have to be adjusted thought the spike cutting after the multiblip fit.  However the current 
    implementation makes sense in that it is similar to how afterspike currents are implemented.
    '''
    if neuron.threshold_components is None:
        raise Exception('reset should never happen at the beginning of a trace')             
        
    tcs = neuron.threshold_components #for ease of updating
    
    # note that these values are at the indicie of the time of the spike which is the index right after the voltage crosses 
    # threshold since the neuron.threshold_components are updated by the dynamics method which is called before the reset. 
    th_spike=tcs['spike'][-1] #this needs to decay through the spike must be very particular about how many indicies to decay
    th_voltage= tcs['voltage'][-1]
    
    # calculate spike component decay though spike from time =1 (not zero because zero is already in neuron.threshold_components
    # via the dynamics method) though the end of the spike cutting
    spike_comp_decay=spike_component_of_threshold_exact(th_spike, b_spike, np.arange(1,neuron.spike_cut_length+1)*neuron.dt) #Note that the plus one is that one needs to know the decay and the inital condition for next starting point 
    
    #update neuron.threshold_components via pass by reference.
    [tcs['voltage'].append(value) for value in np.ones(neuron.spike_cut_length)*th_voltage] #note that here I don't need the plus one because I am starting from zero
    [tcs['spike'].append(value) for value in spike_comp_decay]
    
    # add the amplitude of the spike component decay to last value of vector (reseting)
    tcs['spike'][-1]=tcs['spike'][-1]+a_spike
    
    return tcs['spike'][-1] + tcs['voltage'][-1] + neuron.th_inf * neuron.coeffs['th_inf']


#: The METHOD_LIBRARY constant groups dynamics and reset methods by group name (e.g. 'voltage_dynamics_method').  
#Those groups assign each method in this file a string name.  This is used by the GlifNeuron when initializing 
#its dynamics and reset methods.
METHOD_LIBRARY = {
    'AScurrent_dynamics_method': { 
        'exp':    dynamics_AScurrent_exp,
        'none':   dynamics_AScurrent_none
        },
    'voltage_dynamics_method': { 
        'linear_forward_euler': dynamics_voltage_linear_forward_euler
        },
    'threshold_dynamics_method': {
        'spike_component':        dynamics_threshold_spike_component, 
        'inf':                    dynamics_threshold_inf,
        'three_components_exact': dynamics_threshold_three_components_exact 
        },
    'AScurrent_reset_method': {
        'sum':  reset_AScurrent_sum,
        'none': reset_AScurrent_none
        }, 
    'voltage_reset_method': {
        'v_before':   reset_voltage_v_before,
        'zero':       reset_voltage_zero
        }, 
    'threshold_reset_method': {
        'inf':              reset_threshold_inf,
        'three_components': reset_threshold_three_components
        }
}



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import patch, mock_open, Mock
from simplejson.scanner import JSONDecodeError
import allensdk.core.json_utilities as ju
from allensdk.core.json_utilities import JsonComments
import logging
try:
    import __builtin__ as builtins  # @UnresolvedImport
except:
    import builtins  # @UnresolvedImport


@pytest.fixture
def commented_json():
    return ("{\n"
            "    // comment\n"
            "    \"color\": \"blue\"\n"
            "}")


@pytest.fixture
def blank_line_json():
    return ("{\n"
            "\n"
            "\n"
            "\n"
            "    \"color\": \"blue\"\n"
            "}")


@pytest.fixture
def multi_line_json():
    return ("{\n"
            "/* \n"
            " * multiline comment\n"
            " */\n"
            "    \"color\": \"blue\"\n"
            "}")


@pytest.fixture
def two_multi_line_json():
    return ("{\n"
            "    \"colors\": [\"blue\",\n"
            "    /* comment these out\n"
            "    \"red\",\n"
            "    \"yellow\",\n"
            "     ... but not these */\n"
            "    \"orange\",\n"
            "    \"purple\",\n"
            "    /* also comment this out\n"
            "    \"indigo\",\n"
            "    .... end comment */\n"
            "    \"violet\"\n"
            "    ]\n"
            "}")


@pytest.fixture
def corrupted_json():
    return ("{\n"
            "    \"colors\": \"blue\",\n"
            "    /* comment these out\n"
            "    \"red\",\n"
            "    \"yel")


@pytest.fixture
def ju_logger():
    log = logging.getLogger('allensdk.core.json_utilities')
    log.error = Mock()

    return log


def testSingleLineCommentJSONDecodeError(corrupted_json,
                                    ju_logger):
    with pytest.raises(JSONDecodeError) as e_info:
        with patch(builtins.__name__ + ".open",
                   mock_open(read_data=corrupted_json)):
            JsonComments.read_file("corrupted.json")

    ju_logger.error.assert_called_once_with(
        'Could not load json object from file: corrupted.json')
    assert e_info.typename == 'JSONDecodeError'


def testSingleLineComment(commented_json):
    parsed_json = JsonComments.read_string(
        commented_json)

    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')


def testBlankLines(blank_line_json):
    parsed_json = JsonComments.read_string(
        blank_line_json)

    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')


def testMultiLineComment(multi_line_json):
    parsed_json = JsonComments.read_string(
        multi_line_json)

    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')


def testTwoMultiLineComments(two_multi_line_json):
    parsed_json = JsonComments.read_string(
        two_multi_line_json)

    assert('colors' in parsed_json)
    assert(len(parsed_json['colors']) == 4)
    assert('blue' in parsed_json['colors'])
    assert('orange' in parsed_json['colors'])
    assert('purple' in parsed_json['colors'])
    assert('violet' in parsed_json['colors'])


def testSingleLineCommentFile(commented_json):
    with patch(builtins.__name__ + ".open",
               mock_open(
                   read_data=commented_json)):
        parsed_json = JsonComments.read_file('mock.json')

    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')


def testBlankLinesFile(blank_line_json):
    with patch(builtins.__name__ + ".open",
               mock_open(
                   read_data=blank_line_json)):
        parsed_json = JsonComments.read_file('mock.json')

    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')


def testMultiLineFile(multi_line_json):
    with patch(builtins.__name__ + ".open",
               mock_open(
                   read_data=multi_line_json)):
        parsed_json = JsonComments.read_file('mock.json')

    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')


def testTwoMultiLineFile(two_multi_line_json):
    with patch(builtins.__name__ + ".open",
               mock_open(
                   read_data=two_multi_line_json)):
        parsed_json = JsonComments.read_file('mock.json')

    assert('colors' in parsed_json)
    assert(len(parsed_json['colors']) == 4)
    assert('blue' in parsed_json['colors'])
    assert('orange' in parsed_json['colors'])
    assert('purple' in parsed_json['colors'])
    assert('violet' in parsed_json['colors'])


def test_write_nan():
    with patch(builtins.__name__ + ".open",
               mock_open(),
               create=True) as mo:
        ju.write('/some/file/test.json', { "thing": float('nan')})
    
    assert 'null' in str(mo().write.call_args_list[0])
